import unittest

import numpy as np
import numpy.testing as npt

import wisdem.drivetrainse.generator_models as gm


class TestGenerators(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        self.inputs["machine_rating"] = 5e6
        self.inputs["rated_torque"] = 4.143289e6

        self.inputs["rho_Fe"] = 7700.0
        self.inputs["rho_Fes"] = 7850.0
        self.inputs["rho_Copper"] = 8900.0
        self.inputs["rho_PM"] = 7450.0

        self.inputs["B_r"] = 1.2
        self.inputs["E"] = 2e11
        self.inputs["G"] = 79.3e9
        self.inputs["P_Fe0e"] = 1.0
        self.inputs["P_Fe0h"] = 4.0
        self.inputs["S_N"] = -0.002
        self.inputs["alpha_p"] = 0.5 * np.pi * 0.7
        self.inputs["b_r_tau_r"] = 0.45
        self.inputs["b_ro"] = 0.004
        self.inputs["b_s_tau_s"] = 0.45
        self.inputs["b_so"] = 0.004
        self.inputs["cofi"] = 0.85
        self.inputs["freq"] = 60
        self.inputs["h_i"] = 0.001
        self.inputs["h_sy0"] = 0.0
        self.inputs["h_w"] = 0.005
        self.inputs["k_fes"] = 0.9
        self.inputs["k_fillr"] = 0.7
        self.inputs["k_fills"] = 0.65
        self.inputs["k_s"] = 0.2
        self.discrete_inputs["m"] = 3
        self.inputs["mu_0"] = np.pi * 4e-7
        self.inputs["mu_r"] = 1.06
        self.inputs["p"] = 3.0
        self.inputs["phi"] = np.deg2rad(90)
        self.discrete_inputs["q1"] = 6
        self.discrete_inputs["q2"] = 4
        self.inputs["ratio_mw2pp"] = 0.7
        self.inputs["resist_Cu"] = 1.8e-8 * 1.4
        self.inputs["sigma"] = 40e3
        self.inputs["v"] = 0.3
        self.inputs["y_tau_p"] = 1.0
        self.inputs["y_tau_pr"] = 10.0 / 12
        for k in self.inputs:
            self.inputs[k] = np.array( [self.inputs[k]] )

    def testPMSG_Outer(self):
        myobj = gm.PMSG_Outer(n_pc=20)

        self.inputs["machine_rating"] = 10.321e6
        self.inputs["rated_torque"] = 10.25e6  # rev 1 9.94718e6
        self.inputs["P_mech"] = 10.71947704e6  # rev 1 9.94718e6
        self.inputs["shaft_rpm"] = np.linspace(2, 10, 20)
        self.inputs["rad_ag"] = 4.0  # rev 1  4.92
        self.inputs["len_s"] = 1.7  # rev 2.3
        self.inputs["h_s"] = 0.7  # rev 1 0.3
        self.inputs["p"] = 70  # 100.0    # rev 1 160
        self.inputs["h_m"] = 0.005  # rev 1 0.034
        self.inputs["h_ys"] = 0.04  # rev 1 0.045
        self.inputs["h_yr"] = 0.06  # rev 1 0.045
        self.inputs["b"] = 2.0
        self.inputs["c"] = 5.0
        self.inputs["B_tmax"] = 1.9
        self.inputs["E_p"] = 3300 / np.sqrt(3)
        self.inputs["D_nose"] = 2 * 1.1  # Nose outer radius
        self.inputs["D_shaft"] = 2 * 1.34  # Shaft outer radius =(2+0.25*2+0.3*2)*0.5
        self.inputs["t_r"] = 0.05  # Rotor disc thickness
        self.inputs["h_sr"] = 0.04  # Rotor cylinder thickness
        self.inputs["t_s"] = 0.053  # Stator disc thickness
        self.inputs["h_ss"] = 0.04  # Stator cylinder thickness
        self.inputs["y_sh"] = 0.0005 * 0  # Shaft deflection
        self.inputs["theta_sh"] = 0.00026 * 0  # Slope at shaft end
        self.inputs["y_bd"] = 0.0005 * 0  # deflection at bedplate
        self.inputs["theta_bd"] = 0.00026 * 0  # Slope at bedplate end
        self.inputs["u_allow_pcent"] = 8.5  # % radial deflection
        self.inputs["y_allow_pcent"] = 1.0  # % axial deflection
        self.inputs["z_allow_deg"] = 0.05  # torsional twist
        self.inputs["sigma"] = 60.0e3  # Shear stress
        self.inputs["B_r"] = 1.279
        self.inputs["ratio_mw2pp"] = 0.8
        self.inputs["h_0"] = 5e-3
        self.inputs["h_w"] = 4e-3
        self.inputs["k_fes"] = 0.8
        for k in self.inputs:
            if isinstance(self.inputs[k], float):
                self.inputs[k] = np.array( [self.inputs[k]] )

        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        npt.assert_almost_equal(self.outputs["R_out"], 4.105)
        npt.assert_almost_equal(self.outputs["K_rad"], 0.2125)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio"], 6.24779759)
        npt.assert_almost_equal(self.outputs["tau_p"], 0.17951958)
        npt.assert_almost_equal(self.outputs["tau_s"], 0.14930045)
        npt.assert_almost_equal(self.outputs["b_s"], 0.11203948)
        npt.assert_almost_equal(self.outputs["b_t"], 0.03652435)
        npt.assert_almost_equal(self.outputs["h_t"], 0.709)
        npt.assert_almost_equal(self.outputs["b_m"], 0.14361566)
        npt.assert_almost_equal(self.outputs["B_g"], 0.56284833)
        npt.assert_almost_equal(self.outputs["B_symax"], 0.66753916)
        npt.assert_almost_equal(self.outputs["B_rymax"], 0.44502611)
        npt.assert_almost_equal(self.outputs["B_pm1"], 0.46480944)
        npt.assert_almost_equal(self.outputs["B_smax"], 0.23570673147616897)
        npt.assert_almost_equal(self.outputs["f"][-1], 11.66666667)
        npt.assert_almost_equal(self.outputs["I_s"][-1], 1721.4993685170261)
        npt.assert_almost_equal(self.outputs["R_s"][-1], 0.0862357)
        npt.assert_almost_equal(self.outputs["L_s"], 0.01000474)
        npt.assert_almost_equal(self.outputs["S"], 168.0)
        npt.assert_almost_equal(self.outputs["N_s"], 361.0)
        npt.assert_almost_equal(self.outputs["A_Cuscalc"], 389.46615218)
        npt.assert_almost_equal(self.outputs["J_actual"][-1], 3.125519048273891)
        npt.assert_almost_equal(self.outputs["A_1"][-1], 148362.95007673657)
        npt.assert_almost_equal(len(self.outputs["eandm_efficiency"]), 20)
        npt.assert_almost_equal(self.outputs["eandm_efficiency"][-1], 0.9543687904168252)
        npt.assert_almost_equal(self.outputs["Iron"], 89073.14254723)
        npt.assert_almost_equal(self.outputs["mass_PM"], 1274.81620149)
        npt.assert_almost_equal(self.outputs["Copper"], 13859.17179278)
        npt.assert_almost_equal(self.outputs["twist_r"], 0.00032341)
        npt.assert_almost_equal(self.outputs["twist_s"], 5.8057978e-05)
        npt.assert_almost_equal(self.outputs["Structural_mass"], 62323.08483264)
        npt.assert_almost_equal(self.outputs["generator_mass"], 166530.21537414)

    def testPMSG_Arms(self):
        myobj = gm.PMSG_Arms(n_pc=20)

        self.inputs["shaft_rpm"] = np.linspace(5, 12.1, 20)
        self.inputs["sigma"] = 40e3
        self.inputs["rad_ag"] = 3.26
        self.inputs["len_s"] = 1.60
        self.inputs["h_s"] = 0.070
        self.inputs["tau_p"] = 0.080
        self.inputs["h_m"] = 0.009
        self.inputs["h_ys"] = 0.075
        self.inputs["h_yr"] = 0.075
        self.inputs["n_s"] = 5.0
        self.inputs["b_st"] = 0.480
        self.inputs["n_r"] = 5.0
        self.inputs["b_arm"] = 0.530
        self.inputs["d_r"] = 0.700
        self.inputs["d_s"] = 0.350
        self.inputs["t_d"] = 0.105
        self.inputs["t_wr"] = 0.06
        self.inputs["t_ws"] = 0.06
        self.inputs["D_shaft"] = 2 * 0.43
        self.discrete_inputs["q1"] = 1
        for k in self.inputs:
            if isinstance(self.inputs[k], float):
                self.inputs[k] = np.array( [self.inputs[k]] )

        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        npt.assert_almost_equal(self.outputs["B_symax"], 0.31869052515138424)
        npt.assert_almost_equal(self.outputs["B_tmax"], 1.391188727217009)
        npt.assert_almost_equal(self.outputs["B_smax"], 0.0995499652108797)
        npt.assert_almost_equal(self.outputs["B_rymax"], 0.28682147263624586)
        npt.assert_almost_equal(self.outputs["B_pm1"], 0.674462389329781)
        npt.assert_almost_equal(self.outputs["B_g"], 0.7651537999693548)
        npt.assert_almost_equal(self.outputs["N_s"], 256.0)
        npt.assert_almost_equal(self.outputs["b_s"], 0.012001865684417254)
        npt.assert_almost_equal(self.outputs["b_t"], 0.01466894694762109)
        npt.assert_almost_equal(self.outputs["A_Cuscalc"], 253.53941258331452)
        npt.assert_almost_equal(self.outputs["b_m"], 0.055999999999999994)
        npt.assert_almost_equal(self.outputs["E_p"][-1], 2013.9466358469522)
        npt.assert_almost_equal(self.outputs["f"][-1], 25.813333333333333)
        npt.assert_almost_equal(self.outputs["I_s"][-1], 846.2065921065657)
        npt.assert_almost_equal(self.outputs["R_s"], 0.09770712863768102)
        npt.assert_almost_equal(self.outputs["L_s"], 0.011401106062847888)
        npt.assert_almost_equal(self.outputs["A_1"][-1], 63455.62872651724)
        npt.assert_almost_equal(self.outputs["J_s"][-1], 3.337574160500578)
        npt.assert_almost_equal(self.outputs["Losses"][-1], 351360.55854497873)
        npt.assert_almost_equal(self.outputs["K_rad"], 0.245398773006135)
        npt.assert_almost_equal(len(self.outputs["eandm_efficiency"]), 20)
        npt.assert_almost_equal(self.outputs["eandm_efficiency"][-1], 0.9343418267745142)
        npt.assert_almost_equal(self.outputs["S"], 768.0)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio"], 5.832426544390113)
        npt.assert_almost_equal(self.outputs["Copper"], 6654.6915566955695)
        npt.assert_almost_equal(self.outputs["Iron"], 52673.28069829149)
        npt.assert_almost_equal(self.outputs["TC1"], 16.4856231252069)
        npt.assert_almost_equal(self.outputs["R_out"], 3.3680954773869343)
        npt.assert_almost_equal(self.outputs["Structural_mass"], 33718.05538799999)
        npt.assert_almost_equal(self.outputs["generator_mass"], 94729.99806753898)
        npt.assert_almost_equal(self.outputs["mass_PM"], 1683.970424551947)

    def testPMSG_disc(self):
        myobj = gm.PMSG_Disc(n_pc=20)

        self.inputs["shaft_rpm"] = np.linspace(5, 12.1, 20)
        self.inputs["sigma"] = 40e3
        self.inputs["rad_ag"] = 3.49
        self.inputs["len_s"] = 1.5
        self.inputs["h_s"] = 0.06
        self.inputs["tau_p"] = 0.07
        self.inputs["h_m"] = 0.0105
        self.inputs["h_ys"] = 0.085
        self.inputs["h_yr"] = 0.055
        self.inputs["n_s"] = 5.0
        self.inputs["b_st"] = 0.460
        self.inputs["t_d"] = 0.105
        self.inputs["d_s"] = 0.350
        self.inputs["t_ws"] = 0.150
        self.inputs["D_shaft"] = 2 * 0.43
        self.discrete_inputs["q1"] = 1
        for k in self.inputs:
            if isinstance(self.inputs[k], float):
                self.inputs[k] = np.array( [self.inputs[k]] )

        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        npt.assert_almost_equal(self.outputs["B_symax"], 0.2552951)
        npt.assert_almost_equal(self.outputs["B_smax"], 0.0687816)
        npt.assert_almost_equal(self.outputs["B_tmax"], 1.4426429)
        npt.assert_almost_equal(self.outputs["B_rymax"], 0.3550922)
        npt.assert_almost_equal(self.outputs["B_pm1"], 0.6994079)
        npt.assert_almost_equal(self.outputs["B_g"], 0.7934536)
        npt.assert_almost_equal(self.outputs["N_s"], 314.0000000)
        npt.assert_almost_equal(self.outputs["b_s"], 0.0104753)
        npt.assert_almost_equal(self.outputs["b_t"], 0.0128032)
        npt.assert_almost_equal(self.outputs["A_Cuscalc"], 187.2461758)
        npt.assert_almost_equal(self.outputs["b_m"], 0.0490000)
        npt.assert_almost_equal(self.outputs["E_p"][-1], 2555.3443951)
        npt.assert_almost_equal(self.outputs["f"][-1], 31.6616667)
        npt.assert_almost_equal(self.outputs["I_s"][-1], 657.7809841)
        npt.assert_almost_equal(self.outputs["R_s"], 0.1504414)
        npt.assert_almost_equal(self.outputs["L_s"], 0.0120779)
        npt.assert_almost_equal(self.outputs["A_1"][-1], 56514.1132206)
        npt.assert_almost_equal(self.outputs["J_s"][-1], 3.5129208)
        npt.assert_almost_equal(self.outputs["Losses"][-1], 338164.5549178)
        npt.assert_almost_equal(self.outputs["K_rad"], 0.2148997)
        npt.assert_almost_equal(len(self.outputs["eandm_efficiency"]), 20)
        npt.assert_almost_equal(self.outputs["eandm_efficiency"][-1], 0.936651530)
        npt.assert_almost_equal(self.outputs["S"], 942.0000000)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio"], 5.7277538)
        npt.assert_almost_equal(self.outputs["Copper"], 5588.6107806)
        npt.assert_almost_equal(self.outputs["Iron"], 48400.5614350)
        npt.assert_almost_equal(self.outputs["TC1"], 16.4856231)
        npt.assert_almost_equal(self.outputs["R_out"], 3.6073317)
        npt.assert_almost_equal(self.outputs["Structural_mass"], 67726.0753421)
        npt.assert_almost_equal(self.outputs["generator_mass"], 123674.5978407)
        npt.assert_almost_equal(self.outputs["mass_PM"], 1959.3502831)

    def testEESG(self):
        myobj = gm.EESG(n_pc=20)

        self.inputs["shaft_rpm"] = np.linspace(5, 12.1, 20)
        self.inputs["sigma"] = 48.373e3
        self.inputs["rad_ag"] = 3.2
        self.inputs["len_s"] = 1.4
        self.inputs["h_s"] = 0.060
        self.inputs["tau_p"] = 0.170
        self.inputs["I_f"] = 69
        self.inputs["N_f"] = 100
        self.inputs["h_ys"] = 0.130
        self.inputs["h_yr"] = 0.120
        self.inputs["n_s"] = 5
        self.inputs["b_st"] = 0.470
        self.inputs["n_r"] = 5
        self.inputs["b_arm"] = 0.480
        self.inputs["d_r"] = 0.510
        self.inputs["d_s"] = 0.400
        self.inputs["t_wr"] = 0.140
        self.inputs["t_ws"] = 0.070
        self.inputs["D_shaft"] = 2 * 0.43
        self.discrete_inputs["q1"] = 2
        for k in self.inputs:
            if isinstance(self.inputs[k], float):
                self.inputs[k] = np.array( [self.inputs[k]] )

        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        npt.assert_almost_equal(self.outputs["B_symax"], 0.45940641371625035)
        npt.assert_almost_equal(self.outputs["B_tmax"], 1.8877594585933197)
        npt.assert_almost_equal(self.outputs["B_rymax"], 0.6181179999135469)
        npt.assert_almost_equal(self.outputs["B_g"], 1.1036753874387688)
        npt.assert_almost_equal(self.outputs["N_s"], 236.0)
        npt.assert_almost_equal(self.outputs["b_s"], 0.012779359946805938)
        npt.assert_almost_equal(self.outputs["b_t"], 0.015619217712762814)
        npt.assert_almost_equal(self.outputs["A_Cuscalc"], 228.43105904915618)
        npt.assert_almost_equal(self.outputs["f"][-1], 11.898333333333333)
        npt.assert_almost_equal(self.outputs["I_s"][-1], 858.2107484477335)
        npt.assert_almost_equal(self.outputs["R_s"], 0.0906017600502669)
        npt.assert_almost_equal(self.outputs["A_1"][-1], 60440.40365229799)
        npt.assert_almost_equal(self.outputs["J_s"][-1], 3.75697924800609)
        npt.assert_almost_equal(self.outputs["Losses"][-1], 444556.7203870894)
        npt.assert_almost_equal(self.outputs["K_rad"], 0.21874999999999997)
        npt.assert_almost_equal(len(self.outputs["eandm_efficiency"]), 20)
        npt.assert_almost_equal(self.outputs["eandm_efficiency"][-1], 0.918348408655116)
        npt.assert_almost_equal(self.outputs["S"], 708.0)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio"], 4.695070821210912)
        npt.assert_almost_equal(self.outputs["Copper"], 16059.414075335235)
        npt.assert_almost_equal(self.outputs["Iron"], 72360.88485924648)
        npt.assert_almost_equal(self.outputs["TC1"], 13.632086598066605)
        npt.assert_almost_equal(self.outputs["R_out"], 3.219748743718593)
        npt.assert_almost_equal(self.outputs["Structural_mass"], 42403.44234)
        npt.assert_almost_equal(self.outputs["generator_mass"], 130823.74127458173)

    def testSCIG(self):
        myobj = gm.SCIG(n_pc=20)

        self.inputs["shaft_rpm"] = 100 * np.linspace(2, 12, 20)
        self.inputs["cofi"] = 0.9
        self.inputs["y_tau_p"] = 12.0 / 15.0
        self.inputs["sigma"] = 21.5e3
        self.inputs["rad_ag"] = 0.55
        self.inputs["len_s"] = 1.30
        self.inputs["h_s"] = 0.090
        self.inputs["I_0"] = 140
        self.inputs["B_symax"] = 1.4
        self.discrete_inputs["q1"] = 6
        self.inputs["h_0"] = 0.05
        for k in self.inputs:
            if isinstance(self.inputs[k], float):
                self.inputs[k] = np.array( [self.inputs[k]] )

        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        npt.assert_almost_equal(self.outputs["B_tsmax"], 1.684969767820678)
        npt.assert_almost_equal(self.outputs["B_trmax"], 1.8541950805798264)
        npt.assert_almost_equal(self.outputs["B_rymax"], 1.4)
        npt.assert_almost_equal(self.outputs["B_g"], 0.9267333723013729)
        npt.assert_almost_equal(self.outputs["B_g1"], 0.8617526608763629)
        npt.assert_almost_equal(self.outputs["h_ys"], 0.12135794161089408)
        npt.assert_almost_equal(self.outputs["b_s"], 0.014398966328953218)
        npt.assert_almost_equal(self.outputs["b_t"], 0.017598736624276155)
        npt.assert_almost_equal(self.outputs["D_ratio"], 1.393378075656171)
        npt.assert_almost_equal(self.outputs["D_ratio_UL"], 1.4)
        npt.assert_almost_equal(self.outputs["D_ratio_LL"], 1.37)
        npt.assert_almost_equal(self.outputs["A_Cuscalc"], 244.78242759220464)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio1"], 6.250448674154435)
        npt.assert_almost_equal(self.outputs["h_yr"], 0.12135794161089408)
        npt.assert_almost_equal(self.outputs["tau_p"], 0.5759586531581288)
        npt.assert_almost_equal(self.outputs["b_r"], 0.021513941784534422)
        npt.assert_almost_equal(self.outputs["b_trmin"], 0.023894990015160978)
        npt.assert_almost_equal(self.outputs["b_tr"], 0.026294817736653178)
        npt.assert_almost_equal(self.outputs["rad_r"], 0.547848028863988)
        npt.assert_almost_equal(self.outputs["A_bar"], 0.000968127380304049)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio2"], 2.3240743375043973)
        npt.assert_almost_equal(self.outputs["E_p"][-1], 3584.6363848734995)
        npt.assert_almost_equal(self.outputs["I_s"][-1], 383.73299436233305)
        npt.assert_almost_equal(self.outputs["A_1"][-1], 23985.033858413553)
        npt.assert_almost_equal(self.outputs["J_s"][-1], 1.5676492717917363)
        npt.assert_almost_equal(self.outputs["J_r"][-1], 0.3690452555761286)
        npt.assert_almost_equal(self.outputs["R_s"], 0.00820627176284264)
        npt.assert_almost_equal(self.outputs["R_R"], 0.026060741300117953)
        npt.assert_almost_equal(self.outputs["L_s"], 0.001078616172170861)
        npt.assert_almost_equal(self.outputs["L_sm"], 0.06791816664758084)
        npt.assert_almost_equal(self.outputs["generator_mass"], 40729.31598698678)
        npt.assert_almost_equal(self.outputs["K_rad"], 1.1818181818181817)
        npt.assert_almost_equal(self.outputs["Losses"][-1], 75068.46569051298)
        npt.assert_almost_equal(len(self.outputs["eandm_efficiency"]), 20)
        npt.assert_almost_equal(self.outputs["eandm_efficiency"][-1], 0.9849562794756211)
        npt.assert_almost_equal(self.outputs["Copper"], 1360.253987369221)
        npt.assert_almost_equal(self.outputs["Iron"], 11520.904698026085)
        npt.assert_almost_equal(self.outputs["R_out"], 0.7663579416108941)
        npt.assert_almost_equal(self.outputs["Structural_mass"], 27848.15730159148)
        npt.assert_almost_equal(self.outputs["TC1"], 0.29453832454167966)

    def testDFIG(self):
        myobj = gm.DFIG(n_pc=20)

        self.inputs["shaft_rpm"] = 100 * np.linspace(2, 12, 20)
        self.inputs["cofi"] = 0.9
        self.inputs["y_tau_p"] = 12.0 / 15.0
        self.inputs["sigma"] = 21.5e3
        self.inputs["rad_ag"] = 0.61
        self.inputs["len_s"] = 0.49
        self.inputs["h_s"] = 0.08
        self.inputs["I_0"] = 40.0
        self.inputs["B_symax"] = 1.3
        self.inputs["S_Nmax"] = -0.2
        self.inputs["k_fillr"] = 0.55
        self.inputs["h_0"] = 0.01
        self.inputs["k_fillr"] = 0.55
        self.inputs["k_fills"] = 0.65
        self.discrete_inputs["q1"] = 5
        self.discrete_inputs["q2"] = 4
        for k in self.inputs:
            if isinstance(self.inputs[k], float):
                self.inputs[k] = np.array( [self.inputs[k]] )

        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        npt.assert_almost_equal(self.outputs["B_g"], 1.1073196940043413)
        npt.assert_almost_equal(self.outputs["B_g1"], 1.0207571181956336)
        npt.assert_almost_equal(self.outputs["B_rymax"], 1.3)
        npt.assert_almost_equal(self.outputs["B_tsmax"], 2.0133085345533477)
        npt.assert_almost_equal(self.outputs["B_trmax"], 2.046984459493319)
        npt.assert_almost_equal(self.outputs["h_ys"], 0.1731961572673457)
        npt.assert_almost_equal(self.outputs["b_s"], 0.019163715186897738)
        npt.assert_almost_equal(self.outputs["b_t"], 0.023422318561763904)
        npt.assert_almost_equal(self.outputs["D_ratio"], 1.4232723889628618)
        npt.assert_almost_equal(self.outputs["A_Cuscalc"], 287.4557278034661)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio1"], 4.1745558843775825)
        npt.assert_almost_equal(self.outputs["h_yr"], 0.1731961572673457)
        npt.assert_almost_equal(self.outputs["tau_p"], 0.6387905062299246)
        npt.assert_almost_equal(self.outputs["N_r"], 148.0)
        npt.assert_almost_equal(self.outputs["b_r"], 0.023870136274726762)
        npt.assert_almost_equal(self.outputs["b_trmin"], 0.02869464545814538)
        npt.assert_almost_equal(self.outputs["b_tr"], 0.02917461100244382)
        npt.assert_almost_equal(self.outputs["A_Curcalc"], 5.322395250445834)
        npt.assert_almost_equal(self.outputs["Slot_aspect_ratio2"], 0.41893351109971694)
        npt.assert_almost_equal(self.outputs["E_p"][-1], 1480.0153123457287)
        npt.assert_almost_equal(self.outputs["I_s"][-1], 959.242520719389)
        npt.assert_almost_equal(self.outputs["A_1"][-1], 45049.62948091099)
        npt.assert_almost_equal(self.outputs["J_s"][-1], 3.3370095911785915)
        npt.assert_almost_equal(self.outputs["J_r"][-1], 176.31699359563828)
        npt.assert_almost_equal(self.outputs["R_s"], 0.0037830355111420555)
        npt.assert_almost_equal(self.outputs["R_R"], 0.03168834286109277)
        npt.assert_almost_equal(self.outputs["L_r"], 0.00016404053494963115)
        npt.assert_almost_equal(self.outputs["L_s"], 0.0001855520925628366)
        npt.assert_almost_equal(self.outputs["L_sm"], 0.019568172828126774)
        npt.assert_almost_equal(self.outputs["generator_mass"], 19508.405570203206)
        npt.assert_almost_equal(self.outputs["K_rad"], 0.4016393442622951)
        npt.assert_almost_equal(self.outputs["Losses"][-1], 143901.77088462992)
        npt.assert_almost_equal(len(self.outputs["eandm_efficiency"]), 20)
        npt.assert_almost_equal(self.outputs["eandm_efficiency"][-1], 0.9654635749876888)
        npt.assert_almost_equal(self.outputs["Copper"], 354.97950026786026)
        npt.assert_almost_equal(self.outputs["Iron"], 6077.944572132577)
        npt.assert_almost_equal(self.outputs["Structural_mass"], 13075.48149780277)
        npt.assert_almost_equal(self.outputs["R_out"], 0.8681961572673457)
        npt.assert_almost_equal(self.outputs["TC1"], 0.24790308982258036)
        npt.assert_almost_equal(self.outputs["Current_ratio"][-1], 0.17760183748148742)


if __name__ == "__main__":
    unittest.main()



__author__ = ["Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = ["jake.nunemaker@nrel.gov"]


import os
import re
import datetime as dt
import collections.abc as collections
from copy import deepcopy
from math import ceil
from numbers import Number
from itertools import product

import numpy as np
import pandas as pd
from benedict import benedict

import wisdem.orbit
from wisdem.orbit.phases import DesignPhase, InstallPhase
from wisdem.orbit.core.library import initialize_library, export_library_specs, extract_library_data
from wisdem.orbit.phases.design import (
    SparDesign,
    MonopileDesign,
    ArraySystemDesign,
    ExportSystemDesign,
    MooringSystemDesign,
    ScourProtectionDesign,
    SemiSubmersibleDesign,
    CustomArraySystemDesign,
    OffshoreSubstationDesign,
)
from wisdem.orbit.phases.install import (
    JacketInstallation,
    TurbineInstallation,
    MonopileInstallation,
    MooredSubInstallation,
    ArrayCableInstallation,
    ExportCableInstallation,
    GravityBasedInstallation,
    MooringSystemInstallation,
    ScourProtectionInstallation,
    FloatingSubstationInstallation,
    OffshoreSubstationInstallation,
)
from wisdem.orbit.core.exceptions import PhaseNotFound, WeatherProfileError, PhaseDependenciesInvalid


class ProjectManager:
    """Base Project Manager Class."""

    date_format_short = "%m/%d/%Y"
    date_format_long = "%m/%d/%Y %H:%M"

    _design_phases = (
        MonopileDesign,
        ArraySystemDesign,
        CustomArraySystemDesign,
        ExportSystemDesign,
        ScourProtectionDesign,
        OffshoreSubstationDesign,
        MooringSystemDesign,
        SemiSubmersibleDesign,
        SparDesign,
    )

    _install_phases = (
        MonopileInstallation,
        TurbineInstallation,
        OffshoreSubstationInstallation,
        ArrayCableInstallation,
        ExportCableInstallation,
        ScourProtectionInstallation,
        MooredSubInstallation,
        MooringSystemInstallation,
        GravityBasedInstallation,
        FloatingSubstationInstallation,
        JacketInstallation,
    )

    def __init__(self, config, library_path=None, weather=None):
        """
        Creates and instance of ProjectManager.

        Parameters
        ----------
        config : dict
            Project configuration.
        library_path: str, default: None
            The absolute path to the project library.
        weather : np.ndarray
            Site weather timeseries.
        """

        initialize_library(library_path)
        config = deepcopy(config)
        config = extract_library_data(
            config,
            additional_keys=[
                *config.get("design_phases", []),
                *config.get("install_phases", []),
            ],
        )
        self._phases = {}
        self.config = benedict(config)
        self.resolve_project_capacity()
        self.weather = self.transform_weather_input(weather)

        self.design_results = {}
        self.detailed_outputs = {}

        self.system_costs = {}
        self.installation_costs = {}

        # TODO: Revise:
        self.phase_starts = {}
        self.phase_times = {}
        self._output_logs = []

    def run(self, **kwargs):
        """
        Main project run method.

        Parameters
        ----------
        self.config['design_phases'] : list
            Defines which design phases are ran. These phases are ran before
            install phases and merge the result of the design into self.config.
        self.config['install_phases'] : list | dict
            Defines which installation phases are ran.

            - If ``self.config['install_phases']`` is a list, phases are ran
              sequentially using ``self.run_multiple_phases_in_serial()``.
            - If ``self.config['install_phases']`` is a dict, phases are ran
              using ``self.run_multiple_phases_overlapping()``. The expected
              format for the dictionary is ``{'phase_name': '%m/%d/%Y'}``.
        """

        design_phases = self.config.get("design_phases", [])
        install_phases = self.config.get("install_phases", [])

        if isinstance(design_phases, str):
            design_phases = [design_phases]
        if isinstance(install_phases, str):
            install_phases = [install_phases]

        self.run_all_design_phases(design_phases, **kwargs)

        if isinstance(install_phases, (list, set)):
            self.run_multiple_phases_in_serial(install_phases, **kwargs)

        elif isinstance(install_phases, dict):
            self.run_multiple_phases_overlapping(install_phases, **kwargs)

        if install_phases:
            self.progress = ProjectProgress(self.progress_logs)

        self._print_warnings()

    def _print_warnings(self):
        try:
            df = pd.DataFrame(self.logs)
            df = df.loc[~df["message"].isnull()]
            df = df.loc[df["message"].str.contains("Exceeded")]

            for msg in df["message"].unique():
                idx = df.loc[df["message"] == msg].index[0]
                phase = df.loc[idx, "phase"]
                print(f"{phase}:\n\t {msg}")

        except KeyError:
            pass

    @property
    def phases(self):
        """Returns dict of phases that have been ran."""

        return self._phases

    @classmethod
    def register_design_phase(cls, phase):
        """
        Add a custom design phase to the `ProjectManager` class.

        Parameters
        ----------
        phase : ORBIT.phases.DesignPhase
        """

        if not issubclass(phase, DesignPhase):
            raise ValueError("Registered design phase must be a subclass of " "'ORBIT.phases.DesignPhase'.")

        if phase.__name__ in [c.__name__ for c in cls._design_phases]:
            raise ValueError(f"A phase with name '{phase.__name__}' already exists.")

        if len(re.split("[_ ]", phase.__name__)) > 1:
            raise ValueError(f"Registered phase name must not include a '_'.")

        cls._design_phases = (*cls._design_phases, phase)

    @classmethod
    def register_install_phase(cls, phase):
        """
        Add a custom install phase to the `ProjectManager` class.

        Parameters
        ----------
        phase : ORBIT.phases.InstallPhase
        """

        if not issubclass(phase, InstallPhase):
            raise ValueError("Registered install phase must be a subclass of " "'ORBIT.phases.InstallPhase'.")

        if phase.__name__ in [c.__name__ for c in cls._install_phases]:
            raise ValueError(f"A phase with name '{phase.__name__}' already exists.")

        if len(re.split("[_ ]", phase.__name__)) > 1:
            raise ValueError(f"Registered phase name must not include a '_'.")

        cls._install_phases = (*cls._install_phases, phase)

    @property
    def _capex_categories(self):
        """Returns CapEx categories for phases in `self._install_phases`."""

        out = {}
        for p in self._install_phases:
            try:
                out[p.__name__] = p.capex_category

            except AttributeError:
                out[p.__name__] = "Misc."

        return out

    @property
    def project_params(self):
        """Returns defined project parameters, if found."""

        return self.config.get("project_parameters", {})

    @classmethod
    def compile_input_dict(cls, phases):
        """
        Returns a compiled input dictionary given a list of phases to run.

        Parameters
        ----------
        phases : list
            A collection of offshore design or installation phases.
        """

        _phases = {p: cls.find_key_match(p) for p in phases}
        _error = [n for n, c in _phases.items() if not bool(c)]
        if _error:
            raise PhaseNotFound(_error)

        design_phases = {n: c for n, c in _phases.items() if issubclass(c, DesignPhase)}
        install_phases = {n: c for n, c in _phases.items() if issubclass(c, InstallPhase)}

        config = {}
        for i in install_phases.values():
            config = cls.merge_dicts(config, i.expected_config)

        for d in design_phases.values():
            config = cls.merge_dicts(config, d.expected_config)
            config = cls.remove_keys(config, d.output_config)

        config["project_parameters"] = {
            "turbine_capex": "$/kW (optional, default: 1300)",
            "ncf": "float (optional, default: 0.4)",
            "offtake_price": "$/MWh (optional, default: 80)",
            "project_lifetime": "yrs (optional, default: 25)",
            "discount_rate": "yearly (optional, default: .025)",
            "opex_rate": "$/kW/year (optional, default: 150)",
            "construction_insurance": "$/kW (optional, default: 44)",
            "construction_financing": "$/kW (optional, default: 183)",
            "contingency": "$/kW (optional, default: 316)",
            "commissioning": "$/kW (optional, default: 44)",
            "decommissioning": "$/kW (optional, default: 58)",
            "site_auction_price": "$ (optional, default: 100e6)",
            "site_assessment_cost": "$ (optional, default: 50e6)",
            "construction_plan_cost": "$ (optional, default: 1e6)",
            "installation_plan_cost": "$ (optional, default: 0.25e6)",
        }

        config["design_phases"] = [*design_phases.keys()]
        config["install_phases"] = [*install_phases.keys()]
        config["orbit_version"] = str(wisdem.orbit.__version__)

        return config

    def resolve_project_capacity(self):
        """
        Resolves the relationship between 'project_capacity', 'num_turbines'
        and 'turbine_rating' and verifies that input and calculated values
        match. Adds missing values that can be calculated to the 'self.config'.
        """

        try:
            project_capacity = self.config["plant"]["capacity"]
        except KeyError:
            project_capacity = None

        try:
            turbine_rating = self.config["turbine"]["turbine_rating"]
        except KeyError:
            turbine_rating = None

        try:
            num_turbines = self.config["plant"]["num_turbines"]
        except KeyError:
            num_turbines = None

        if all((project_capacity, turbine_rating, num_turbines)):
            if project_capacity != (turbine_rating * num_turbines):
                raise AttributeError(f"Input and calculated project capacity don't match.")

        else:
            if all((project_capacity, turbine_rating)):
                num_turbines = ceil(project_capacity / turbine_rating)
                self.config["plant"]["num_turbines"] = num_turbines

            elif all((project_capacity, num_turbines)):
                turbine_rating = project_capacity / num_turbines
                try:
                    self.config["turbine"]["turbine_rating"] = turbine_rating

                except KeyError:
                    self.config["turbine"] = {"turbine_rating": turbine_rating}

            elif all((num_turbines, turbine_rating)):
                project_capacity = turbine_rating * num_turbines
                self.config["plant"]["capacity"] = project_capacity

    @classmethod
    def find_key_match(cls, target):
        """
        Searches cls.phase_dict() for a key that matches text in 'target'.

        Parameters
        ----------
        target : str
            Phase name to search for a match with.

        Returns
        -------
        phase_class : BasePhase | None
            Matched module class or None if no match is found.
        """

        phase = re.split("[_ ]", target)[0]
        phase_dict = cls.phase_dict()
        phase_class = phase_dict.get(phase, None)

        return phase_class

    @classmethod
    def phase_dict(cls):
        """
        Returns dictionary of all possible phases with format 'name': 'class'.
        """

        install = {p.__name__: p for p in cls._install_phases}
        design = {p.__name__: p for p in cls._design_phases}

        return {**install, **design}

    @classmethod
    def merge_dicts(cls, left, right, overwrite=True, add_keys=True):
        """
        Merges two dicts (right into left) with an option to add keys of right.

        Parameters
        ----------
        left : dict
        right : dict
        add_keys : bool

        Returns
        -------
        new : dict
            Merged dictionary.
        """
        new = left.copy()
        if not add_keys:
            right = {k: right[k] for k in set(new).intersection(set(right))}

        for k, _ in right.items():
            if k in new and isinstance(new[k], dict) and isinstance(right[k], collections.Mapping):
                new[k] = cls.merge_dicts(new[k], right[k], overwrite=overwrite, add_keys=add_keys)
            elif k in new and isinstance(new[k], list) and isinstance(right[k], list):
                new[k].extend(right[k])
            else:
                if overwrite or k not in new:
                    new[k] = right[k]

                else:
                    continue

        return new

    @classmethod
    def remove_keys(cls, left, right):
        """
        Recursively removes keys from left that are found in right.

        Parameters
        ----------
        left : dict
        right : dict

        Returns
        -------
        new : dict
            Left dictionary with keys of right removed.
        """

        new = left.copy()
        right = {k: right[k] for k in set(new).intersection(set(right))}

        for k, val in right.items():
            if isinstance(new.get(k, None), dict) and isinstance(val, dict):
                new[k] = cls.remove_keys(new[k], val)

                if not new[k]:
                    _ = new.pop(k)

            else:
                _ = new.pop(k)

        return new

    def create_config_for_phase(self, phase):
        """
        Produces a configuration input dictionary for 'phase'.

        This method will pick the most specific definition of each parameter.
        For example, if self.master_config['site']['distance'] and
        self.config['PhaseName']['site']['distance'] are both defined,
        the latter will be chosen as it is more specific. This allows for phase
        specific definitions, eg. distance to port dependent on phase.

        Parameters
        ----------
        phase : str
            Name of phase. Phase specific information will be pulled from
            self.config['PhaseName'] if this key exists.

        Returns
        -------
        phase_config : dict
            Configuration dictionary with phase specific information merged in.
        """

        _specific = self.config.get(phase, {}).copy()
        _general = {k: v for k, v in self.config.items() if k not in set(self.phase_dict())}

        phase_config = self.merge_dicts(_general, _specific)

        return phase_config

    @property
    def phase_ends(self):
        ret = {}
        for k, t in self.phase_times.items():
            try:
                ret[k] = self.phase_starts[k] + t

            except KeyError:
                pass

        return ret

    def run_install_phase(self, name, start, **kwargs):
        """
        Compiles the phase specific configuration input dictionary for input
        'name', checks the input against _class.expected_config and runs the
        phase calculations with 'phase.run()'.

        Parameters
        ----------
        name : str
            Phase to run.
        weather : None | np.ndarray

        Returns
        -------
        time : int | float
            Total phase time.
        logs : list
            List of phase logs.
        """

        if self.weather is not None:
            weather = self.get_weather_profile(start)

        else:
            weather = None

        _catch = kwargs.get("catch_exceptions", False)
        _class = self.get_phase_class(name)
        _config = self.create_config_for_phase(name)
        processes = _config.pop("processes", {})
        kwargs = {**kwargs, **processes}

        if _catch:
            try:
                phase = _class(_config, weather=weather, phase_name=name, **kwargs)
                phase.run()

            except Exception as e:
                print(f"\n\t - {name}: {e}")
                return None, None

        else:
            phase = _class(_config, weather=weather, phase_name=name, **kwargs)
            phase.run()

        self._phases[name] = phase

        time = phase.total_phase_time
        logs = deepcopy(phase.env.logs)

        self.phase_starts[name] = start
        self.phase_times[name] = time
        self.detailed_outputs = self.merge_dicts(self.detailed_outputs, phase.detailed_output)

        if phase.system_capex:
            self.system_costs[name] = phase.system_capex

        if phase.installation_capex:
            self.installation_costs[name] = phase.installation_capex

        return time, logs

    def get_phase_class(self, phase):
        """
        Returns the class object for input 'phase'.

        Parameters
        ----------
        phase : str
            Name of phase. Must match a class name in either
            'self._install_phases' or 'self._design_phases'.

        Returns
        -------
        phase_class : Phase
            Class of base type Phase that represents input 'phase'.
        """

        _dict = self.phase_dict()
        phase_class = self.find_key_match(phase)
        if phase_class is None:
            raise PhaseNotFound(phase)

        return phase_class

    def run_all_design_phases(self, phase_list, **kwargs):
        """
        Runs multiple design phases and adds '.design_result' to self.config.
        """

        for name in phase_list:
            self.run_design_phase(name, **kwargs)

    def run_design_phase(self, name, **kwargs):
        """
        Runs a design phase defined by 'name' and merges the '.design_result'
        into self.config.

        Parameters
        ----------
        name : str
            Name of design phase that partially matches a key in `phase_dict`.
        """

        _catch = kwargs.get("catch_exceptions", False)
        _class = self.get_phase_class(name)
        _config = self.create_config_for_phase(name)

        if _catch:
            try:
                phase = _class(_config)
                phase.run()

            except Exception as e:
                print(f"\n\t - {name}: {e}")
                return

        else:
            phase = _class(_config)
            phase.run()

        self._phases[name] = phase

        self.design_results = self.merge_dicts(self.design_results, phase.design_result, overwrite=False)

        self.config = self.merge_dicts(self.config, phase.design_result, overwrite=False)
        self.detailed_outputs = self.merge_dicts(self.detailed_outputs, phase.detailed_output)

    def run_multiple_phases_in_serial(self, phase_list, **kwargs):
        """
        Runs multiple phases listed in self.config['install_phases'] in serial.

        Parameters
        ----------
        phase_list : list
            List of installation phases to run.
        """

        start = 0

        for name in phase_list:
            time, logs = self.run_install_phase(name, start, **kwargs)

            if logs is None:
                continue

            else:
                for l in logs:
                    try:
                        l["time"] += start
                    except KeyError:
                        pass

                self._output_logs.extend(logs)
                start = ceil(start + time)

    def run_multiple_phases_overlapping(self, phases, **kwargs):
        """
        Runs multiple phases overlapping using a mixture of dates, indices or
        dependencies.

        Parameters
        ----------
        phases : dict
            Dictionary of phases to run.
        """

        defined, variable = self._parse_install_phase_values(phases)
        zero = min(defined.values())

        # Run defined
        for name, start in defined.items():
            _, logs = self.run_install_phase(name, start, **kwargs)

            if logs is None:
                continue

            else:
                for l in logs:
                    try:
                        l["time"] += start - zero
                    except KeyError:
                        pass

                self._output_logs.extend(logs)

        # Run remaining phases
        self.run_dependent_phases(variable, zero, **kwargs)

    def run_dependent_phases(self, _phases, zero, **kwargs):
        """
        Runs remaining phases that depend on other phase times.

        Parameters
        ----------
        _phases : dict
            Dictionary of phases to run.
        zero : int | float
            Zero time for the simulation. Used to aggregate total logs.
        """

        phases = deepcopy(_phases)
        skipped = {}

        while True:
            phases = {**phases, **skipped}
            if not phases:
                break

            skipped = {}
            progress = False

            for name in list(phases.keys()):
                target, perc = phases.pop(name)

                try:
                    start = self.get_dependency_start_time(target, perc)
                    _, logs = self.run_install_phase(name, start, **kwargs)

                    progress = True

                    if logs is None:
                        continue

                    else:
                        for l in logs:
                            try:
                                l["time"] += start - zero
                            except KeyError:
                                pass

                        self._output_logs.extend(logs)

                except KeyError:
                    skipped[name] = (target, perc)

            if not progress:
                raise PhaseDependenciesInvalid(_phases)

    def get_dependency_start_time(self, target, perc):
        """
        Returns start time based on the `perc` complete of `target` phase.

        Parameters
        ----------
        target : str
            Phase that start time is dependent on.
        perc : int | float
            Percentage of the target phase completion time. `0`: starts at the
            same time. `1`: starts when target phase is completed.
        """

        start = self.phase_starts[target]
        elapsed = self.phase_times[target]

        return start + elapsed * perc

    @staticmethod
    def transform_weather_input(weather):
        """
        Checks that an input weather profile matches the required format and
        converts the index to a datetime index if necessary.

        Parameters
        ----------
        weather : pd.DataFrame
        """

        if weather is None:
            return None

        else:
            try:
                weather = weather.set_index("datetime")
                if not isinstance(weather.index, pd.DatetimeIndex):
                    weather.index = pd.to_datetime(weather.index)

            except KeyError:
                pass

            return weather

    def _parse_install_phase_values(self, phases):
        """
        Parses the input dictionary `install_phases`, splitting them into
        phases that have defined start times and ones that rely on other phases.

        Parameters
        ----------
        phases : dict
            Dictionary of installation phases to run.

        Raises
        ------
        ValueError
            Raised if no phases have a defined start date as the project can't
            be tied to a specific part of the weather profile.
        """

        defined = {}
        depends = {}

        for k, v in phases.items():
            if isinstance(v, (int, float)):
                defined[k] = ceil(v)

            elif isinstance(v, str):
                _dt = dt.datetime.strptime(v, self.date_format_short)

                try:
                    i = self.weather.index.get_loc(_dt)
                    defined[k] = i

                except AttributeError:
                    raise ValueError(f"No weather profile configured " f"for '{k}': '{v}' input type.")

                except KeyError:
                    raise WeatherProfileError(_dt, self.weather)

            elif isinstance(v, tuple) and len(v) == 2:
                depends[k] = v

            else:
                raise ValueError(f"Input type '{k}': '{v}' not recognized.")

        if not defined:
            raise ValueError("No phases have a defined start index/date.")

        return defined, depends

    def get_weather_profile(self, start):
        """
        Pulls weather profile from 'self.weather' starting at 'start', raising
        any errors if needed.

        Parameters
        ----------
        start : datetime
            Starting index for output weather profile.

        Returns
        -------
        profile : np.ndarray.
            Weather profile with first index at 'start'.
        """

        return self.weather.iloc[ceil(start) :].copy().to_records()

    def outputs(self, include_logs=False, npv_detailed=False):
        """Returns dict of all available outputs."""

        out = {
            # Times
            "project_time": self.project_time,
            "installation_time": self.installation_time,
            # Costs
            "capex_breakdown": self.capex_breakdown,
            "capex_breakdown_per_kw": self.capex_breakdown_per_kw,
            "turbine_capex": self.turbine_capex,
            "turbine_capex_per_kw": self.turbine_capex_per_kw,
            "installation_capex": self.installation_capex,
            "installation_capex_per_kw": self.installation_capex_per_kw,
            "system_capex": self.system_capex,
            "system_capex_per_kw": self.system_capex_per_kw,
            "overnight_capex": self.overnight_capex,
            "overnight_capex_per_kw": self.overnight_capex_per_kw,
            "soft_capex": self.soft_capex,
            "soft_capex_per_kw": self.soft_capex_per_kw,
            "bos_capex": self.bos_capex,
            "bos_capex_per_kw": self.bos_capex_per_kw,
            "project_capex": self.project_capex,
            "project_capex_per_kw": self.project_capex_per_kw,
            "total_capex": self.total_capex,
            "total_capex_per_kw": self.total_capex_per_kw,
            "npv": self.npv,
        }

        if include_logs:
            out["logs"] = self.actions

        if npv_detailed:
            out = {
                **out,
                **{
                    "cash_flow": self.cash_flow,
                    "monthly_revenue": self.monthly_revenue,
                    "monthly_expenses": self.monthly_expenses,
                },
            }

        return out

    @property
    def capacity(self):
        """Returns project capacity in MW."""

        try:
            capacity = self.config["plant"]["capacity"]

        except KeyError:
            capacity = None

        return capacity

    @property
    def num_turbines(self):
        """Returns number of turbines in the project."""

        try:
            num_turbines = self.config["plant"]["num_turbines"]

        except KeyError:
            num_turbines = None

        return num_turbines

    @property
    def turbine_rating(self):
        """Returns turbine rating in MW"""

        try:
            rating = self.config["turbine"]["turbine_rating"]

        except KeyError:
            rating = None

        return rating

    @property
    def logs(self):
        """Returns list of all logs in the project."""

        return sorted(self._output_logs, key=lambda l: l["time"])

    @property
    def project_time(self):
        """Returns total project time as the time of the last log."""

        return self.actions[-1]["time"]

    @property
    def month_bins(self):
        """Returns bins representing project months."""

        return np.arange(0, self.project_time + 730, 730)

    @property
    def monthly_expenses(self):
        """Returns the monthly expenses of the project from development through
        construction."""

        opex = self.monthly_opex
        lifetime = self.project_params.get("project_lifetime", 25)

        _expense_logs = self._filter_logs(keys=["cost", "time"])
        expenses = np.array(_expense_logs, dtype=[("cost", "f8"), ("time", "i4")])
        dig = np.digitize(expenses["time"], self.month_bins)

        monthly = {}
        for i in range(1, lifetime * 12):
            monthly[i] = sum(expenses["cost"][dig == i]) + opex[i]

        return monthly

    @property
    def monthly_opex(self):
        """Returns the monthly OpEx expenditures based on project size."""

        rate = self.project_params.get("opex_rate", 150)
        lifetime = self.project_params.get("project_lifetime", 25)

        try:
            times, turbines = self.progress.energize_points
            dig = list(np.digitize(times, self.month_bins))

        except ValueError:
            return {i: 0.0 for i in range(1, lifetime * 12)}

        opex = {}
        for i in range(1, lifetime * 12):
            generating_strings = len([t for t in dig if i >= t])
            generating_turbines = sum(turbines[:generating_strings])

            opex[i] = generating_turbines * self.turbine_rating * rate * 1000 / 12

        return opex

    @property
    def monthly_revenue(self):
        """Returns the monthly revenue based on when array system strings can
        be energized, eg. 'self.progress.energize_points'."""

        ncf = self.project_params.get("ncf", 0.4)
        price = self.project_params.get("offtake_price", 80)
        lifetime = self.project_params.get("project_lifetime", 25)

        times, turbines = self.progress.energize_points
        dig = list(np.digitize(times, self.month_bins))

        revenue = {}
        for i in range(1, lifetime * 12):
            generating_strings = len([t for t in dig if i >= t])
            generating_turbines = sum(turbines[:generating_strings])
            production = generating_turbines * self.turbine_rating * ncf * 730  # MWh
            revenue[i] = production * price

        return revenue

    @property
    def cash_flow(self):
        """Returns the net cash flow based on `self.monthly_expenses` and
        `self.monthly_revenue`."""

        try:
            revenue = self.monthly_revenue

        except ValueError:
            revenue = {}

        expenses = self.monthly_expenses
        return {
            i: revenue.get(i, 0.0) - expenses.get(i, 0.0)
            for i in range(1, max([*revenue.keys(), *expenses.keys()]) + 1)
        }

    @property
    def npv(self):
        """Returns the net present value of the project based on
        `self.cash_flow`."""

        dr = self.project_params.get("discount_rate", 0.025)
        pr = (1 + dr) ** (1 / 12) - 1

        cash_flow = self.cash_flow
        _npv = [(cash_flow[i] / (1 + pr) ** (i)) for i in range(1, max(cash_flow.keys()) + 1)]

        return (self.total_capex - self.installation_capex) - sum(_npv)

    @property
    def progress_logs(self):
        """Returns logs of progress points."""

        return self._filter_logs(keys=["progress", "time"])

    def _filter_logs(self, keys):
        """Returns filtered list of logs."""

        filtered = []
        for l in self.logs:
            try:
                filtered.append(tuple(l[k] for k in keys))

            except KeyError:
                pass

        return filtered

    @property
    def progress_summary(self):
        """Returns a summary of progress by month."""

        arr = np.array(self.progress_logs, dtype=[("progress", "U32"), ("time", "i4")])
        dig = np.digitize(arr["time"], self.month_bins)

        summary = {}
        for i in range(1, len(self.month_bins)):
            unique, counts = np.unique(arr["progress"][dig == i], return_counts=True)
            summary[i] = dict(zip(unique, counts))

        return summary

    @property
    def actions(self):
        """Returns list of all actions in the project."""

        actions = [l for l in self.logs if l["level"] == "ACTION"]
        return sorted(actions, key=lambda l: l["time"])

    @staticmethod
    def create_input_xlsx():
        """
        A wrapper around self.compile_input_dict that produces an excel input
        file instead of a .json file.
        """
        pass

    @property
    def phase_dates(self):
        """
        Returns a combination of input start dates and `self.phase_times`.
        """

        if not isinstance(self.config["install_phases"], dict):
            print("Project was not configured with start dates.")
            return None

        dates = {}

        for phase, _start in self.config["install_phases"].items():
            start = dt.datetime.strptime(_start, self.date_format_short)
            end = start + dt.timedelta(hours=ceil(self.phase_times[phase]))

            dates[phase] = {
                "start": start.strftime(self.date_format_long),
                "end": end.strftime(self.date_format_long),
            }

        return dates

    @property
    def installation_time(self):
        """
        Returns sum of installation module times. This does not consider
        overlaps if phase dates are supplied.
        """

        res = sum(
            [v for k, v in self.phase_times.items() if k in self.config["install_phases"] and isinstance(v, Number)]
        )
        return res

    @property
    def project_days(self):
        """
        Returns days elapsed during installation phases accounting for
        overlapping phases.
        """

        dates = self.phase_dates
        starts = [d["start"] for _, d in dates.items()]
        ends = [d["end"] for _, d in dates.items()]
        return max([self._diff_dates_long(*p) for p in product(starts, ends)])

    def _diff_dates_long(self, a, b):
        """Returns the difference of two dates in `self.date_format_long`."""

        if not isinstance(a, dt.datetime):
            a = dt.datetime.strptime(a, self.date_format_long)

        if not isinstance(b, dt.datetime):
            b = dt.datetime.strptime(b, self.date_format_long)

        return abs((a - b).days)

    @property
    def overnight_capex_per_kw(self):
        """
        Returns overnight CAPEX/kW.
        """

        try:
            capex = self.overnight_capex / (self.capacity * 1000)

        except TypeError:
            capex = None

        return capex

    @property
    def system_capex(self):
        """Returns total system procurement CapEx."""

        return np.nansum([c for _, c in self.system_costs.items()])

    @property
    def system_capex_per_kw(self):
        """Returns system CapEx/kW."""

        try:
            capex = self.system_capex / (self.capacity * 1000)

        except TypeError:
            capex = None

        return capex

    @property
    def installation_capex(self):
        """Returns total installation related CapEx."""

        return np.nansum([c for _, c in self.installation_costs.items()])

    @property
    def installation_capex_per_kw(self):
        """Returns installation CapEx/kW."""

        try:
            capex = self.installation_capex / (self.capacity * 1000)

        except TypeError:
            capex = None

        return capex

    @property
    def capex_breakdown(self):
        """Returns CapEx breakdown by category."""

        unique = np.unique([*self.system_costs.keys(), *self.installation_costs.keys()])
        categories = {}

        for phase in unique:
            for base, cat in self._capex_categories.items():
                if base in phase:
                    categories[phase] = cat
                    break

        missing = [p for p in unique if p not in categories.keys()]
        if missing:
            print(f"Warning: CapEx category not found for {missing}. " f"Added to 'Misc.'")

            for phase in missing:
                categories[phase] = "Misc."

        outputs = {}
        for phase, cost in self.system_costs.items():
            name = categories[phase]
            if name in outputs.keys():
                outputs[name] += cost

            else:
                outputs[name] = cost

        for phase, cost in self.installation_costs.items():
            name = categories[phase] + " Installation"
            if name in outputs.keys():
                outputs[name] += cost

            else:
                outputs[name] = cost

        outputs["Turbine"] = self.turbine_capex
        outputs["Soft"] = self.soft_capex
        outputs["Project"] = self.project_capex

        return outputs

    @property
    def capex_breakdown_per_kw(self):
        """Returns CapEx per kW breakdown by category."""

        return {k: v / (self.capacity * 1000) for k, v in self.capex_breakdown.items()}

    @property
    def bos_capex(self):
        """Returns total balance of system CapEx."""

        return self.system_capex + self.installation_capex

    @property
    def bos_capex_per_kw(self):
        """Returns balance of system CapEx/kW."""

        try:
            capex = self.bos_capex / (self.capacity * 1000)

        except TypeError:
            capex = None

        return capex

    @property
    def turbine_capex(self):
        """Returns the total turbine CAPEX."""

        _capex = self.project_params.get("turbine_capex", 1300)
        try:
            num_turbines = self.config["plant"]["num_turbines"]
            rating = self.config["turbine"]["turbine_rating"]

        except KeyError:
            raise KeyError(
                f"Total turbine CAPEX can't be calculated. Required "
                f"parameters 'plant.num_turbines' or 'turbine.turbine_rating' "
                f"not found."
            )

        capex = _capex * num_turbines * rating * 1000
        return capex

    @property
    def turbine_capex_per_kw(self):
        """Returns the turbine CapEx/kW."""

        _capex = self.project_params.get("turbine_capex", 1300)
        return _capex

    @property
    def overnight_capex(self):
        """Returns the overnight capital cost of the project."""

        return self.system_capex + self.turbine_capex

    @property
    def soft_capex(self):
        """Returns total project cost costs."""

        try:
            capex = self.soft_capex_per_kw * self.capacity * 1000

        except TypeError:
            capex = None

        return capex

    @property
    def soft_capex_per_kw(self):
        """
        Returns project soft costs per kW. Default numbers are based on the
        Cost of Energy Review (Stehly and Beiter 2018).
        """

        insurance = self.project_params.get("construction_insurance", 44)
        financing = self.project_params.get("construction_financing", 183)
        contingency = self.project_params.get("contingency", 316)
        commissioning = self.project_params.get("commissioning", 44)
        decommissioning = self.project_params.get("decommissioning", 58)

        return sum([insurance, financing, contingency, commissioning, decommissioning])

    @property
    def project_capex(self):
        """
        Returns project related CapEx line items. To override the defaults,
        the keys below should be passed to the 'project_parameters' subdict.
        """

        site_auction = self.project_params.get("site_auction_price", 100e6)
        site_assessment = self.project_params.get("site_assessment_cost", 50e6)
        construction_plan = self.project_params.get("construction_plan_cost", 1e6)
        installation_plan = self.project_params.get("installation_plan_cost", 0.25e6)

        return sum(
            [
                site_auction,
                site_assessment,
                construction_plan,
                installation_plan,
            ]
        )

    @property
    def project_capex_per_kw(self):
        """Returns project related CapEx per kW."""

        try:
            capex = self.project_capex / (self.capacity * 1000)

        except TypeError:
            capex = None

        return capex

    @property
    def total_capex(self):
        """Returns total project CapEx including soft costs."""

        return self.bos_capex + self.turbine_capex + self.soft_capex + self.project_capex

    @property
    def total_capex_per_kw(self):
        """Returns total CapEx/kW."""

        try:
            capex = self.total_capex / (self.capacity * 1000)

        except TypeError:
            capex = None

        return capex

    def export_configuration(self, file_name):
        """
        Exports the configuration settings for the project to
        `library_path/project/config/file_name.yaml`.

        Parameters
        ----------
        file_name : str
            Name to use for the file.
        """

        export_library_specs("config", file_name, self.config)

    def export_project_logs(self, filepath, level="ACTION"):
        """
        Exports the project logs to a .csv file.

        Parameters
        ----------
        filepath : str
            Filepath to save logs at.
        level : str, optional
            Log level to save.
            Options: 'ACTION' | 'DEBUG'
            Default: 'ACTION'
        """

        dirs = os.path.split(filepath)[0]
        if dirs and not os.path.isdir(dirs):
            os.makedirs(dirs)

        if level == "ACTION":
            out = pd.DataFrame(self.actions)

        elif level == "DEBUG":
            out = pd.DataFrame(self.logs)

        else:
            raise ValueError(f"Unrecognized level '{level}'." " Must be 'ACTION' or 'DEBUG'.")

        out.to_csv(filepath, index=False)


class ProjectProgress:
    """Class to store, parse and return project progress data."""

    def __init__(self, data):
        """
        Creates an instance of `ProjectProgress`.

        Parameters
        ----------
        data : list
            List of tuples representing progress points of the project.
        """

        self.data = data

    @property
    def complete_export_system(self):
        """
        Returns project time when the export system and offshore substations(s)
        installations were completed (max of individual values).
        """

        export = self.parse_logs("Export System")
        substations = self.parse_logs("Offshore Substation")

        return max([*export, *substations])

    @property
    def complete_array_strings(self):
        """
        Returns list of times that the array strings and associated
        substructure/turbine assembly installations were completed.
        """

        strings = self.parse_logs("Array String")
        _subs = self.parse_logs("Substructure")
        _turbines = self.parse_logs("Turbine")

        per_string = len(_turbines) // len(strings)
        if len(_turbines) % len(strings):
            per_string += 1

        subs = self.chunk_max(_subs, per_string)
        turbines = self.chunk_max(_turbines, per_string)
        num_turbines = list(self.chunk_len(_turbines, per_string))

        data = list(zip(strings, subs, turbines))

        return [max(l) for l in data], num_turbines

    @property
    def energize_points(self):
        """
        Returns list of times where an array string can be energized. Max of
        each value in `self.complete_array_strings` and
        `self.complete_export_system`.
        """

        export = self.complete_export_system
        points = []

        times, turbines = self.complete_array_strings
        for t in times:
            points.append(max([t, export]))

        return points, turbines

    def parse_logs(self, k):
        """Parse `self.data` for specific progress points associated key `k`"""

        pts = [p[1] for p in self.data if p[0] == k]
        if not pts:
            raise ValueError(f"Installed '{k}' not found in project logs.")

        return pts

    @staticmethod
    def chunk_max(l, n):
        """Yield max value of successive n-sized chunks from l."""

        for i in range(0, len(l), n):
            yield max(l[i : i + n])

    @staticmethod
    def chunk_len(l, n):
        """Yield successive n-sized chunks from l."""

        for i in range(0, len(l), n):
            yield len(l[i : i + n])



# 1 ---------

from math import pi

import numpy as np
import matplotlib.pyplot as plt

from wisdem.ccblade.ccblade import CCBlade, CCAirfoil

plot_flag = False

# geometry
Rhub = 1.5
Rtip = 63.0

r = np.array(
    [
        2.8667,
        5.6000,
        8.3333,
        11.7500,
        15.8500,
        19.9500,
        24.0500,
        28.1500,
        32.2500,
        36.3500,
        40.4500,
        44.5500,
        48.6500,
        52.7500,
        56.1667,
        58.9000,
        61.6333,
    ]
)
chord = np.array(
    [
        3.542,
        3.854,
        4.167,
        4.557,
        4.652,
        4.458,
        4.249,
        4.007,
        3.748,
        3.502,
        3.256,
        3.010,
        2.764,
        2.518,
        2.313,
        2.086,
        1.419,
    ]
)
theta = np.array(
    [
        13.308,
        13.308,
        13.308,
        13.308,
        11.480,
        10.162,
        9.011,
        7.795,
        6.544,
        5.361,
        4.188,
        3.125,
        2.319,
        1.526,
        0.863,
        0.370,
        0.106,
    ]
)
B = 3  # number of blades

tilt = 5.0
precone = 2.5
yaw = 0.0

nSector = 8  # azimuthal discretization

# atmosphere
rho = 1.225
mu = 1.81206e-5

# power-law wind shear profile
shearExp = 0.2
hubHt = 90.0

# 1 ----------


# 2 ----------
import os
from wisdem.inputs.validation import load_geometry_yaml
baseyaml = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "02_reference_turbines", "nrel5mw.yaml")
data = load_geometry_yaml(baseyaml)
af = data['airfoils']
af_names = ["Cylinder", "Cylinder", "DU40_A17", "DU35_A17", "DU30_A17", "DU25_A17", "DU21_A17", "NACA64_A17"]
airfoil_types = [0] * len(af_names)
for i in range(len(af_names)):
    for j in range(len(af)):
        if af[j]["name"] == af_names[i]:
            polars = af[j]['polars'][0]
            airfoil_types[i] = CCAirfoil(
                np.rad2deg(polars["c_l"]["grid"]),
                [polars["re"]],
                polars["c_l"]["values"],
                polars["c_d"]["values"],
                polars["c_m"]["values"],
            )

# place at appropriate radial stations
af_idx = [0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7]

af = [0] * len(r)
for i in range(len(r)):
    af[i] = airfoil_types[af_idx[i]]

# 2 ----------

# 3 ----------

# create CCBlade object
rotor = CCBlade(r, chord, theta, af, Rhub, Rtip, B, rho, mu, precone, tilt, yaw, shearExp, hubHt, nSector)

# 3 ----------

# 4 ----------

# set conditions
Uinf = 10.0
tsr = 7.55
pitch = 0.0
Omega = Uinf * tsr / Rtip * 30.0 / pi  # convert to RPM
azimuth = 0.0

# evaluate distributed loads
loads, derivs = rotor.distributedAeroLoads(Uinf, Omega, pitch, azimuth)
Np = loads["Np"]
Tp = loads["Tp"]

# 4 ----------

# 5 ----------

# plot
rstar = (r - Rhub) / (Rtip - Rhub)

# append zero at root and tip
rstar = np.concatenate([[0.0], rstar, [1.0]])
Np = np.concatenate([[0.0], Np, [0.0]])
Tp = np.concatenate([[0.0], Tp, [0.0]])

if plot_flag:
    plt.plot(rstar, Tp / 1e3, label="lead-lag")
    plt.plot(rstar, Np / 1e3, label="flapwise")
    plt.xlabel("blade fraction")
    plt.ylabel("distributed aerodynamic loads (kN)")
    plt.legend(loc="upper left")
    plt.grid()
    plt.show()
# 5 ----------

# plt.savefig('/Users/sning/Dropbox/NREL/SysEng/TWISTER/ccblade-beta-setup/docs/images/distributedAeroLoads.pdf')
# plt.savefig('/Users/sning/Dropbox/NREL/SysEng/TWISTER/ccblade-beta-setup/docs/images/distributedAeroLoads.png')

# 6 ----------

outputs, derivs = rotor.evaluate([Uinf], [Omega], [pitch])

P = outputs["P"]
T = outputs["T"]
Q = outputs["Q"]

outputs, derivs = rotor.evaluate([Uinf], [Omega], [pitch], coefficients=True)

CP = outputs["CP"]
CT = outputs["CT"]
CQ = outputs["CQ"]

print("CP =", CP)
print("CT =", CT)
print("CQ =", CQ)

# 6 ----------

# 7 ----------

# velocity has a small amount of Reynolds number dependence
tsr = np.linspace(2, 14, 50)
Omega = 10.0 * np.ones_like(tsr)
Uinf = Omega * pi / 30.0 * Rtip / tsr
pitch = np.zeros_like(tsr)

outputs, derivs = rotor.evaluate(Uinf, Omega, pitch, coefficients=True)

CP = outputs["CP"]
CT = outputs["CT"]
CQ = outputs["CQ"]

if plot_flag:
    plt.figure()
    plt.plot(tsr, CP)
    plt.xlabel("$\lambda$")
    plt.ylabel("$c_p$")
    plt.show()
# 7 ----------

# plt.savefig('/Users/sning/Dropbox/NREL/SysEng/TWISTER/ccblade-beta-setup/docs/images/cp.pdf')
# plt.savefig('/Users/sning/Dropbox/NREL/SysEng/TWISTER/ccblade-beta-setup/docs/images/cp.png')



import numpy as np
import openmdao.api as om

import wisdem.drivetrainse.layout as lay
import wisdem.drivetrainse.drive_structure as ds
import wisdem.drivetrainse.drive_components as dc
from wisdem.drivetrainse.hub import Hub_System
from wisdem.drivetrainse.gearbox import Gearbox
from wisdem.drivetrainse.generator import Generator


class DriveMaterials(om.ExplicitComponent):
    """
    This component sifts through the material database and sets the material property data structures needed in this module
    """

    def initialize(self):
        self.options.declare("n_mat")
        self.options.declare("direct", default=False)

    def setup(self):
        n_mat = self.options["n_mat"]

        self.add_input("E_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("G_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("Xt_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("Xy_mat", val=np.zeros(n_mat), units="Pa")
        self.add_input("wohler_exp_mat", val=np.zeros(n_mat))
        self.add_input("wohler_A_mat", val=np.zeros(n_mat))
        self.add_input("rho_mat", val=np.zeros(n_mat), units="kg/m**3")
        self.add_input("unit_cost_mat", val=np.zeros(n_mat), units="USD/kg")
        self.add_discrete_input("material_names", val=n_mat * [""])
        self.add_discrete_input("lss_material", "steel")
        self.add_discrete_input("hss_material", "steel")
        self.add_discrete_input("hub_material", "iron")
        self.add_discrete_input("spinner_material", "carbon")
        self.add_discrete_input("bedplate_material", "steel")

        self.add_output("hub_E", val=0.0, units="Pa")
        self.add_output("hub_G", val=0.0, units="Pa")
        self.add_output("hub_rho", val=0.0, units="kg/m**3")
        self.add_output("hub_Xy", val=0.0, units="Pa")
        self.add_output("hub_wohler_exp", val=0.0)
        self.add_output("hub_wohler_A", val=0.0)
        self.add_output("hub_mat_cost", val=0.0, units="USD/kg")
        self.add_output("spinner_rho", val=0.0, units="kg/m**3")
        self.add_output("spinner_Xt", val=0.0, units="Pa")
        self.add_output("spinner_mat_cost", val=0.0, units="USD/kg")
        self.add_output("lss_E", val=0.0, units="Pa")
        self.add_output("lss_G", val=0.0, units="Pa")
        self.add_output("lss_rho", val=0.0, units="kg/m**3")
        self.add_output("lss_Xy", val=0.0, units="Pa")
        self.add_output("lss_Xt", val=0.0, units="Pa")
        self.add_output("lss_wohler_exp", val=0.0)
        self.add_output("lss_wohler_A", val=0.0)
        self.add_output("lss_cost", val=0.0, units="USD/kg")
        self.add_output("hss_E", val=0.0, units="Pa")
        self.add_output("hss_G", val=0.0, units="Pa")
        self.add_output("hss_rho", val=0.0, units="kg/m**3")
        self.add_output("hss_Xy", val=0.0, units="Pa")
        self.add_output("hss_Xt", val=0.0, units="Pa")
        self.add_output("hss_wohler_exp", val=0.0)
        self.add_output("hss_wohler_A", val=0.0)
        self.add_output("hss_cost", val=0.0, units="USD/kg")
        self.add_output("bedplate_E", val=0.0, units="Pa")
        self.add_output("bedplate_G", val=0.0, units="Pa")
        self.add_output("bedplate_rho", val=0.0, units="kg/m**3")
        self.add_output("bedplate_Xy", val=0.0, units="Pa")
        self.add_output("bedplate_mat_cost", val=0.0, units="USD/kg")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Convert to isotropic material
        E = np.mean(inputs["E_mat"], axis=1)
        G = np.mean(inputs["G_mat"], axis=1)
        # Take the minimum Xt in longitudinal and transversal diretion, neglect direction 3 (through the fibers)
        Xt = inputs["Xt_mat"][:, [0, 1]].min(axis=1)
        sigy = inputs["Xy_mat"]
        m = inputs["wohler_exp_mat"]
        A = inputs["wohler_A_mat"]
        rho = inputs["rho_mat"]
        cost = inputs["unit_cost_mat"]

        hub_name = discrete_inputs["hub_material"]
        spin_name = discrete_inputs["spinner_material"]
        lss_name = discrete_inputs["lss_material"]
        bed_name = discrete_inputs["bedplate_material"]
        mat_names = discrete_inputs["material_names"]

        # Get the index into the material list
        spin_imat = mat_names.index(spin_name)
        hub_imat = mat_names.index(hub_name)
        lss_imat = mat_names.index(lss_name)
        bed_imat = mat_names.index(bed_name)

        outputs["hub_E"] = E[hub_imat]
        outputs["hub_G"] = G[hub_imat]
        outputs["hub_rho"] = rho[hub_imat]
        outputs["hub_Xy"] = sigy[hub_imat]
        outputs["hub_wohler_exp"] = m[hub_imat]
        outputs["hub_wohler_A"] = A[hub_imat]
        outputs["hub_mat_cost"] = cost[hub_imat]

        outputs["spinner_rho"] = rho[spin_imat]
        outputs["spinner_Xt"] = Xt[spin_imat]
        if Xt[spin_imat] == 0.0:
            raise Exception(
                "The tensile strength of the composite used in the rotor hub spinner is zero. Please check your input file."
            )
        outputs["spinner_mat_cost"] = cost[spin_imat]

        outputs["lss_E"] = E[lss_imat]
        outputs["lss_G"] = G[lss_imat]
        outputs["lss_rho"] = rho[lss_imat]
        outputs["lss_Xy"] = sigy[lss_imat]
        outputs["lss_Xt"] = Xt[lss_imat]
        if Xt[lss_imat] == 0.0:
            raise Exception(
                "The tensile strength of the material used in the low speed shaft is zero. Please check your input file."
            )
        outputs["lss_wohler_exp"] = m[lss_imat]
        outputs["lss_wohler_A"] = A[lss_imat]
        outputs["lss_cost"] = cost[lss_imat]

        outputs["bedplate_E"] = E[bed_imat]
        outputs["bedplate_G"] = G[bed_imat]
        outputs["bedplate_rho"] = rho[bed_imat]
        outputs["bedplate_Xy"] = sigy[bed_imat]
        outputs["bedplate_mat_cost"] = cost[bed_imat]

        if not self.options["direct"]:
            hss_name = discrete_inputs["hss_material"]
            hss_imat = mat_names.index(hss_name)
            outputs["hss_E"] = E[hss_imat]
            outputs["hss_G"] = G[hss_imat]
            outputs["hss_rho"] = rho[hss_imat]
            outputs["hss_Xy"] = sigy[hss_imat]
            outputs["hss_Xt"] = Xt[hss_imat]
            if Xt[hss_imat] == 0.0:
                raise Exception(
                    "The tensile strength of the material used in the high speed shaft is zero. Please check your input file."
                )
            outputs["hss_wohler_exp"] = m[hss_imat]
            outputs["hss_wohler_A"] = A[hss_imat]
            outputs["hss_cost"] = cost[hss_imat]


# ----------------------------------------------------------------------------------------------
class DrivetrainSE(om.Group):
    """
    DirectDriveSE defines an OpenMDAO group that represents a wind turbine drivetrain without a gearbox and two main bearings.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]["WISDEM"]["DriveSE"]
        n_dlcs = self.options["modeling_options"]["WISDEM"]["n_dlc"]
        direct = opt["direct"]
        if direct:
            use_gb_torque_density = False
        else:
            use_gb_torque_density = opt["use_gb_torque_density"]
        dogen = self.options["modeling_options"]["flags"]["generator"]
        n_pc = self.options["modeling_options"]["WISDEM"]["RotorSE"]["n_pc"]

        self.set_input_defaults("machine_rating", units="kW")
        self.set_input_defaults("planet_numbers", [3, 3, 0])
        self.set_input_defaults("gear_configuration", "eep")
        self.set_input_defaults("hvac_mass_coeff", 0.025, units="kg/kW/m")
        # self.set_input_defaults('mb1Type', 'CARB')
        # self.set_input_defaults('mb2Type', 'SRB')
        self.set_input_defaults("uptower", True)
        self.set_input_defaults("upwind", True)
        self.set_input_defaults("n_blades", 3)

        # Materials prep
        self.add_subsystem(
            "mat",
            DriveMaterials(direct=direct, n_mat=self.options["modeling_options"]["materials"]["n_mat"]),
            promotes=["*"],
        )

        # Need to do these first, before the layout
        self.add_subsystem("hub", Hub_System(modeling_options=opt["hub"]), promotes=["*"])
        self.add_subsystem("gear", Gearbox(direct_drive=direct, use_gb_torque_density=use_gb_torque_density), promotes=["*"])

        # Layout and mass for the big items
        if direct:
            self.add_subsystem("layout", lay.DirectLayout(), promotes=["*"])
        else:
            self.add_subsystem("layout", lay.GearedLayout(), promotes=["*"])

        # All the smaller items
        self.add_subsystem("bear1", dc.MainBearing())
        self.add_subsystem("bear2", dc.MainBearing())
        self.add_subsystem("brake", dc.Brake(direct_drive=direct), promotes=["*"])
        self.add_subsystem("elec", dc.Electronics(), promotes=["*"])
        self.add_subsystem("yaw", dc.YawSystem(), promotes=["yaw_mass", "yaw_I", "yaw_cm", "rotor_diameter", "D_top"])

        # Generator
        self.add_subsystem("rpm", dc.RPM_Input(n_pc=n_pc), promotes=["*"])
        if dogen:
            gentype = self.options["modeling_options"]["WISDEM"]["GeneratorSE"]["type"]
            self.add_subsystem(
                "generator",
                Generator(design=gentype, n_pc=n_pc),
                promotes=[
                    "generator_mass",
                    "generator_cost",
                    "generator_I",
                    "machine_rating",
                    "generator_efficiency",
                    "rated_torque",
                    ("rotor_mass", "generator_rotor_mass"),
                    ("rotor_I", "generator_rotor_I"),
                    ("stator_mass", "generator_stator_mass"),
                    ("stator_I", "generator_stator_I"),
                ],
            )
        else:
            self.add_subsystem("gensimp", dc.GeneratorSimple(direct_drive=direct, n_pc=n_pc), promotes=["*"])

        # Final tallying
        self.add_subsystem("misc", dc.MiscNacelleComponents(), promotes=["*"])
        self.add_subsystem("nac", dc.NacelleSystemAdder(direct_drive=direct), promotes=["*"])
        self.add_subsystem("rna", dc.RNA_Adder(), promotes=["*"])

        # Structural analysis
        self.add_subsystem(
            "lss", ds.Hub_Rotor_LSS_Frame(n_dlcs=n_dlcs, modeling_options=opt, direct_drive=direct), promotes=["*"]
        )
        if direct:
            self.add_subsystem(
                "nose", ds.Nose_Stator_Bedplate_Frame(modeling_options=opt, n_dlcs=n_dlcs), promotes=["*"]
            )
        else:
            self.add_subsystem("hss", ds.HSS_Frame(modeling_options=opt, n_dlcs=n_dlcs), promotes=["*"])
            self.add_subsystem("bed", ds.Bedplate_IBeam_Frame(modeling_options=opt, n_dlcs=n_dlcs), promotes=["*"])

        # Dynamics
        self.add_subsystem("dyn", dc.DriveDynamics(), promotes=["*"])

        # Output-to-input connections
        self.connect("bedplate_rho", ["pitch_system.rho", "spinner.metal_rho"])
        self.connect("bedplate_Xy", ["pitch_system.Xy", "spinner.Xy"])
        self.connect("bedplate_mat_cost", "spinner.metal_cost")
        self.connect("hub_rho", ["hub_shell.rho", "rho_castiron"])
        self.connect("hub_Xy", "hub_shell.Xy")
        self.connect("hub_mat_cost", "hub_shell.metal_cost")
        self.connect("spinner_rho", ["spinner.composite_rho", "rho_fiberglass"])
        self.connect("spinner_Xt", "spinner.composite_Xt")
        self.connect("spinner_mat_cost", "spinner.composite_cost")

        if direct:
            self.connect("D_bearing1", "bear1.D_bearing")
            self.connect("D_bearing2", "bear2.D_bearing")

        self.connect("bear1.mb_mass", "mb1_mass")
        self.connect("bear1.mb_I", "mb1_I")
        self.connect("bear1.mb_max_defl_ang", "mb1_max_defl_ang")
        self.connect("s_mb1", "mb1_cm")
        self.connect("bear2.mb_mass", "mb2_mass")
        self.connect("bear2.mb_I", "mb2_I")
        self.connect("bear2.mb_max_defl_ang", "mb2_max_defl_ang")
        self.connect("s_mb2", "mb2_cm")
        self.connect("bedplate_rho", "yaw.rho")
        self.connect("s_gearbox", "gearbox_cm")
        self.connect("s_generator", "generator_cm")

        if dogen:
            self.connect("generator.R_out", "R_generator")
            self.connect("bedplate_E", "generator.E")
            self.connect("bedplate_G", "generator.G")

            if direct:
                self.connect("lss_rpm", "generator.shaft_rpm")
                self.connect("torq_deflection", "generator.y_sh")
                self.connect("torq_angle", "generator.theta_sh")
                self.connect("stator_deflection", "generator.y_bd")
                self.connect("stator_angle", "generator.theta_bd")

                self.linear_solver = lbgs = om.LinearBlockGS()
                self.nonlinear_solver = nlbgs = om.NonlinearBlockGS()
                nlbgs.options["maxiter"] = 3
                nlbgs.options["atol"] = nlbgs.options["atol"] = 1e-2
                nlbgs.options["iprint"] = 0
            else:
                self.connect("hss_rpm", "generator.shaft_rpm")



import unittest

import numpy as np
import numpy.testing as npt

import wisdem.rotorse.rotor_elasticity as rel
from wisdem.commonse.cross_sections import Tube


class TestRE(unittest.TestCase):
    def setUp(self):
        self.discrete_inputs = {}
        self.discrete_outputs = {}
        self.inputs = {}
        options = {}
        opt_options = {}
        options = {}
        options["WISDEM"] = {}
        options["WISDEM"]["RotorSE"] = {}
        options["WISDEM"]["RotorSE"]["n_xy"] = npts = 101
        options["WISDEM"]["RotorSE"]["n_span"] = nspan = 31
        options["WISDEM"]["RotorSE"]["n_webs"] = nweb = 0
        options["WISDEM"]["RotorSE"]["n_layers"] = nlay = 1
        options["WISDEM"]["RotorSE"]["te_ss"] = "none"
        options["WISDEM"]["RotorSE"]["te_ps"] = "none"
        options["WISDEM"]["RotorSE"]["spar_cap_ss"] = "none"
        options["WISDEM"]["RotorSE"]["spar_cap_ps"] = "none"
        options["WISDEM"]["RotorSE"]["layer_name"] = ["mylayer"]
        options["WISDEM"]["RotorSE"]["layer_mat"] = ["mymat"]
        options["materials"] = {}
        options["materials"]["n_mat"] = nmat = 1
        options["General"] = {}
        options["General"]["verbosity"] = False

        # Use a tubular blade
        self.inputs["r"] = np.linspace(0.0, 10.0, nspan)
        self.inputs["theta"] = np.zeros(nspan)
        self.inputs["chord"] = 2 * np.ones(nspan)
        self.inputs["pitch_axis"] = 0.5 * np.ones(nspan) # 0=LE, 1=TE
        self.inputs["precurve"] = np.zeros(nspan)
        self.inputs["presweep"] = np.zeros(nspan)
        self.inputs["coord_xy_interp"] = np.zeros( (nspan, npts, 2) )
        angle = np.linspace(0, 2*np.pi, npts)
        for k in range(nspan):
            self.inputs["coord_xy_interp"][k,:,0] = 0.5 * np.cos(angle) + 0.5
            self.inputs["coord_xy_interp"][k,:,1] = 0.5 * np.sin(angle)
        self.inputs["coord_xy_interp"][:,0,1] = 0.0
        self.inputs["coord_xy_interp"][:,-1,1] = 0.0
        self.inputs["uptilt"] = np.zeros(1)
        self.discrete_inputs["n_blades"] = 3
        self.inputs["web_start_nd"] = self.inputs["web_end_nd"] = np.zeros((nweb, nspan))
        self.inputs["layer_web"] = np.zeros(nlay)
        self.discrete_inputs["definition_layer"] = np.ones(nlay)
        self.inputs["layer_thickness"] = 0.01 * np.ones((nlay, nspan))
        self.inputs["layer_start_nd"] = np.zeros((nlay, nspan))
        self.inputs["layer_end_nd"] = np.ones((nlay, nspan))
        self.inputs["fiber_orientation"] = 90 * np.ones((nlay, nspan))

        self.discrete_inputs["mat_name"] = ["mymat"]
        self.discrete_inputs["orth"] = np.zeros(nmat)
        self.inputs["E"] = 1e10 * np.ones((nmat,3))
        self.inputs["G"] = 1e8 * np.ones((nmat,3))
        self.inputs["nu"] = 0.33 * np.ones((nmat,3))
        self.inputs["rho"] = 1e3 * np.ones(nmat)
        self.inputs["joint_position"] = self.inputs["joint_mass"] = 0.0

        self.myobj = rel.RunPreComp(modeling_options=options, opt_options=opt_options)
        rotorse_options = options["WISDEM"]["RotorSE"]
        self.myobj.n_span = rotorse_options["n_span"]
        self.myobj.n_webs = rotorse_options["n_webs"]
        self.myobj.n_layers = rotorse_options["n_layers"]
        self.myobj.n_xy = rotorse_options["n_xy"]  # Number of coordinate points to describe the airfoil geometry
        mat_init_options = options["materials"]
        self.myobj.n_mat = mat_init_options["n_mat"]
        self.myobj.verbosity = options["General"]["verbosity"]

        self.myobj.te_ss_var = rotorse_options["te_ss"]
        self.myobj.te_ps_var = rotorse_options["te_ps"]
        self.myobj.spar_cap_ss_var = rotorse_options["spar_cap_ss"]
        self.myobj.spar_cap_ps_var = rotorse_options["spar_cap_ps"]

        self.mytube = Tube(self.inputs["chord"][0], self.inputs["layer_thickness"][0,0])
        
    def test_no_pitch(self):
        outputs = {}
        
        self.myobj.compute(self.inputs, outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(outputs["z"], self.inputs["r"])
        npt.assert_almost_equal(outputs["x_cg"], 0.0)
        npt.assert_almost_equal(outputs["x_ec"], 0.0)
        npt.assert_almost_equal(outputs["x_tc"], 0.0)
        npt.assert_almost_equal(outputs["x_sc"], 0.0)
        npt.assert_almost_equal(outputs["y_cg"], 0.0)
        npt.assert_almost_equal(outputs["y_ec"], 0.0)
        npt.assert_almost_equal(outputs["y_tc"], 0.0)
        npt.assert_almost_equal(outputs["y_sc"], 0.0)
        npt.assert_almost_equal(outputs["EIxy"], 0.0, decimal=6)
        npt.assert_almost_equal(outputs["A"], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["EIxx"]/self.inputs["E"][0,0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(outputs["EIyy"]/self.inputs["E"][0,0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(outputs["GJ"]/self.inputs["G"][0,0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(outputs["EA"]/self.inputs["E"][0,0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["rhoA"]/self.inputs["rho"][0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["rhoJ"]/self.inputs["rho"][0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(outputs["Tw_iner"], 0.0)
        npt.assert_almost_equal(outputs["flap_iner"]/self.inputs["rho"][0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(outputs["edge_iner"]/self.inputs["rho"][0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(outputs["blade_mass"], self.mytube.Area*self.inputs["rho"][0]*self.inputs["r"][-1], decimal=-1)
        npt.assert_almost_equal(outputs["mass_all_blades"], 3*outputs["blade_mass"], decimal=3)
        idx = np.int_(np.floor(0.5*self.inputs["r"].size))
        npt.assert_almost_equal(outputs["blade_span_cg"], self.inputs["r"][idx], decimal=1)
        npt.assert_almost_equal(outputs["blade_moment_of_inertia"], outputs["blade_mass"]*self.inputs["r"][-1]**2/3.0, decimal=-1)

        '''
        outputs["xu_spar"]
        outputs["xl_spar"]
        outputs["yu_spar"]
        outputs["yl_spar"]
        outputs["xu_te"]
        outputs["xl_te"]
        outputs["yu_te"]
        outputs["yl_te"]
        outputs["I_all_blades"]
        '''
        
    def test_with_pitch(self):
        outputs = {}
        self.inputs["theta"] = 45 * np.ones(self.inputs["theta"].shape)
        
        self.myobj.compute(self.inputs, outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(outputs["z"], self.inputs["r"])
        npt.assert_almost_equal(outputs["x_cg"], 0.0)
        npt.assert_almost_equal(outputs["x_ec"], 0.0)
        npt.assert_almost_equal(outputs["x_tc"], 0.0)
        npt.assert_almost_equal(outputs["x_sc"], 0.0)
        npt.assert_almost_equal(outputs["y_cg"], 0.0)
        npt.assert_almost_equal(outputs["y_ec"], 0.0)
        npt.assert_almost_equal(outputs["y_tc"], 0.0)
        npt.assert_almost_equal(outputs["y_sc"], 0.0)
        npt.assert_almost_equal(outputs["EIxy"], 0.0, decimal=6)
        npt.assert_almost_equal(outputs["A"], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["EIxx"]/self.inputs["E"][0,0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(outputs["EIyy"]/self.inputs["E"][0,0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(outputs["GJ"]/self.inputs["G"][0,0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(outputs["EA"]/self.inputs["E"][0,0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["rhoA"]/self.inputs["rho"][0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["rhoJ"]/self.inputs["rho"][0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(outputs["Tw_iner"], 45.0)
        npt.assert_almost_equal(outputs["flap_iner"]/self.inputs["rho"][0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(outputs["edge_iner"]/self.inputs["rho"][0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(outputs["blade_mass"], self.mytube.Area*self.inputs["rho"][0]*self.inputs["r"][-1], decimal=-1)
        npt.assert_almost_equal(outputs["mass_all_blades"], 3*outputs["blade_mass"], decimal=3)
        idx = np.int_(np.floor(0.5*self.inputs["r"].size))
        npt.assert_almost_equal(outputs["blade_span_cg"], self.inputs["r"][idx], decimal=1)
        npt.assert_almost_equal(outputs["blade_moment_of_inertia"], outputs["blade_mass"]*self.inputs["r"][-1]**2/3.0, decimal=-1)
        
        
    def test_with_le_pitch_axis(self):
        outputs = {}
        self.inputs["pitch_axis"] = np.zeros(self.inputs["pitch_axis"].shape)
        
        self.myobj.compute(self.inputs, outputs, self.discrete_inputs, self.discrete_outputs)

        npt.assert_equal(outputs["z"], self.inputs["r"])
        npt.assert_almost_equal(outputs["x_cg"], 0.0)
        npt.assert_almost_equal(outputs["x_ec"], 0.0)
        npt.assert_almost_equal(outputs["x_tc"], 0.0)
        npt.assert_almost_equal(outputs["x_sc"], 0.0)
        npt.assert_almost_equal(outputs["y_cg"], 1.0)
        npt.assert_almost_equal(outputs["y_ec"], 0.0)
        npt.assert_almost_equal(outputs["y_tc"], 1.0)
        npt.assert_almost_equal(outputs["y_sc"], 1.0)
        npt.assert_almost_equal(outputs["EIxy"], 0.0, decimal=6)
        npt.assert_almost_equal(outputs["A"], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["EIxx"]/self.inputs["E"][0,0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(outputs["EIyy"]/self.inputs["E"][0,0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(outputs["GJ"]/self.inputs["G"][0,0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(outputs["EA"]/self.inputs["E"][0,0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["rhoA"]/self.inputs["rho"][0], self.mytube.Area, decimal=3)
        npt.assert_almost_equal(outputs["rhoJ"]/self.inputs["rho"][0], self.mytube.J0, decimal=3)
        npt.assert_almost_equal(outputs["Tw_iner"], 0.0)
        npt.assert_almost_equal(outputs["flap_iner"]/self.inputs["rho"][0], self.mytube.Iyy, decimal=3)
        npt.assert_almost_equal(outputs["edge_iner"]/self.inputs["rho"][0], self.mytube.Ixx, decimal=3)
        npt.assert_almost_equal(outputs["blade_mass"], self.mytube.Area*self.inputs["rho"][0]*self.inputs["r"][-1], decimal=-1)
        npt.assert_almost_equal(outputs["mass_all_blades"], 3*outputs["blade_mass"], decimal=3)
        idx = np.int_(np.floor(0.5*self.inputs["r"].size))
        npt.assert_almost_equal(outputs["blade_span_cg"], self.inputs["r"][idx], decimal=1)
        npt.assert_almost_equal(outputs["blade_moment_of_inertia"], outputs["blade_mass"]*self.inputs["r"][-1]**2/3.0, decimal=-1)
        
if __name__ == "__main__":
    unittest.main()



"""Tests for the `MooringSystemDesign` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

import pytest

from wisdem.orbit.phases.design import MooringSystemDesign

base = {
    "site": {"depth": 200},
    "turbine": {"turbine_rating": 6},
    "plant": {"num_turbines": 50},
}


@pytest.mark.parametrize("depth", range(10, 1000, 100))
def test_depth_sweep(depth):
    config = deepcopy(base)
    config["site"]["depth"] = depth

    m = MooringSystemDesign(config)
    m.run()

    assert m.design_result
    assert m.total_cost


@pytest.mark.parametrize("rating", range(3, 15, 1))
def test_rating_sweeip(rating):
    config = deepcopy(base)
    config["turbine"]["turbine_rating"] = rating

    m = MooringSystemDesign(config)
    m.run()

    assert m.design_result
    assert m.total_cost


def test_drag_embedment_fixed_length():
    m = MooringSystemDesign(base)
    m.run()

    baseline = m.line_length

    default = deepcopy(base)
    default["mooring_system_design"] = {"anchor_type": "Drag Embedment"}

    m = MooringSystemDesign(default)
    m.run()

    with_default = m.line_length
    assert with_default > baseline

    custom = deepcopy(base)
    custom["mooring_system_design"] = {
        "anchor_type": "Drag Embedment",
        "drag_embedment_fixed_length": 1000,
    }

    m = MooringSystemDesign(custom)
    m.run()

    assert m.line_length > with_default
    assert m.line_length > baseline


def test_custom_num_lines():
    config = deepcopy(base)
    config["mooring_system_design"] = {"num_lines": 5}

    m = MooringSystemDesign(config)
    m.run()

    assert m.design_result["mooring_system"]["num_lines"] == 5



"""Tests for the `MooredSubInstallation` class and related infrastructure."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import pandas as pd
import pytest

from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.phases.install import MooredSubInstallation
from wisdem.test.test_orbit.data import test_weather

config = extract_library_specs("config", "moored_install")
no_supply = extract_library_specs("config", "moored_install_no_supply")


def test_simulation_setup():
    sim = MooredSubInstallation(config)
    assert sim.config == config
    assert sim.env

    assert sim.support_vessel
    assert len(sim.sub_assembly_lines) == config["port"]["sub_assembly_lines"]
    assert len(sim.turbine_assembly_lines) == config["port"]["turbine_assembly_cranes"]
    assert len(sim.installation_groups) == config["towing_vessel_groups"]["num_groups"]
    assert sim.num_turbines == config["plant"]["num_turbines"]


@pytest.mark.parametrize("weather", (None, test_weather), ids=["no_weather", "test_weather"])
@pytest.mark.parametrize("config", (config, no_supply))
def test_for_complete_logging(weather, config):
    sim = MooredSubInstallation(config, weather=weather)
    sim.run()

    df = pd.DataFrame(sim.env.actions)
    df = df.assign(shift=(df["time"] - df["time"].shift(1)))

    for vessel in df["agent"].unique():
        _df = df[df["agent"] == vessel].copy()
        _df = _df.assign(shift=(_df["time"] - _df["time"].shift(1)))
        assert (_df["shift"] - _df["duration"]).abs().max() < 1e-9

    assert ~df["cost"].isnull().any()
    _ = sim.agent_efficiencies
    _ = sim.detailed_output



import pandas as pd
import pytest
import openmdao.api as om

from wisdem.landbosse.landbosse_omdao.landbosse import LandBOSSE
from wisdem.landbosse.landbosse_omdao.OpenMDAODataframeCache import OpenMDAODataframeCache


@pytest.fixture
def landbosse_costs_by_module_type_operation():
    """
    Executes LandBOSSE and extracts cost output for the regression
    test.
    """
    prob = om.Problem(reports=False)
    prob.model = LandBOSSE()
    prob.setup()
    prob.run_model()
    #prob.model.list_inputs(units=True, desc=False)
    landbosse_costs_by_module_type_operation = prob["landbosse_costs_by_module_type_operation"]
    return landbosse_costs_by_module_type_operation


def compare_expected_to_actual(expected_df, actual_module_type_operation_list, validation_output_csv):
    """
    This compares the expected costs as calculated by a prior model run
    with the actual results from a current model run.

    It compares the results row by row and prints any differences.

    Parameters
    ----------
    expected_df : pd.DataFrame
        The absolute filename of the expected output .xlsx file.

    actual_module_type_operation_list : str
        The module_type_operation_list as returned by a subclass of
        XlsxManagerRunner.

    validation_output_xlsx : str
        The absolute pathname to the output file with the comparison
        results.

    Returns
    -------
    bool
        True if the expected and actual results are equal. It returns
        False otherwise.
    """
    # First, make the list of dictionaries into a dataframe, and drop
    # the raw_cost and raw_cost_total_or_per_turbine columns.
    actual_df = pd.DataFrame(actual_module_type_operation_list)

    columns_to_compare = ["Cost / project", "Project ID with serial", "Module", "Type of cost"]
    cost_per_project_actual = actual_df[columns_to_compare]
    cost_per_project_expected = expected_df[columns_to_compare]

    comparison = cost_per_project_actual.merge(
        cost_per_project_expected, on=["Project ID with serial", "Module", "Type of cost"]
    )

    comparison.rename(
        columns={"Cost / project_x": "Cost / project actual", "Cost / project_y": "Cost / project expected"},
        inplace=True,
    )

    comparison["% delta"] = (comparison["Cost / project actual"] / comparison["Cost / project expected"] - 1) * 100
    comparison.to_csv(validation_output_csv, index=False)

    # If the comparison dataframe is empty, that means there are no common
    # projects in the expected data that match the actual data.
    if len(comparison) < 1:
        print("=" * 80)
        print("Validation error: There are no common projects between actual and expected data.")
        print("=" * 80)
        return False

    # Find all rows where the difference is unequal to 0. These are rows
    # that failed validation. Note that, after the join, the rows may be
    # in a different order than the originals.
    #
    # Round the difference to a given number of decimal places.
    failed_rows = comparison[~pd.isnull(comparison["% delta"]) & comparison["% delta"].round(decimals=4) != 0]

    if len(failed_rows) > 0:
        print("=" * 80)
        print("The following rows failed validation:")
        print(failed_rows)
        print("=" * 80)
        return False
    else:
        return True


def test_landbosse(landbosse_costs_by_module_type_operation):
    """
    This runs the regression test by comparing against the expected validation
    data.
    """
    OpenMDAODataframeCache._cache = {}  # Clear the cache
    expected_validation_data_sheets = OpenMDAODataframeCache.read_all_sheets_from_xlsx("ge15_expected_validation")
    costs_by_module_type_operation = expected_validation_data_sheets["costs_by_module_type_operation"]
    result = compare_expected_to_actual(
        costs_by_module_type_operation, landbosse_costs_by_module_type_operation, "test.csv"
    )

    assert result

    




def skipLines(f, n):
    for i in range(n):
        f.readline()

class Orthotropic2DMaterial:
    """Represents a homogeneous orthotropic material in a
    plane stress state.

    """

    def __init__(self, E1, E2, G12, nu12, rho, name=""):
        """a struct-like object.  all inputs are also fields.
        The object also has an identification
        number *.mat_idx so unique materials can be identified.

        Parameters
        ----------
        E1 : float (N/m^2)
            Young's modulus in first principal direction
        E2 : float (N/m^2)
            Young's modulus in second principal direction
        G12 : float (N/m^2)
            shear modulus
        nu12 : float
            Poisson's ratio  (nu12*E22 = nu21*E11)
        rho : float (kg/m^3)
            density
        name : str
            an optional identifier

        """
        self.E1 = E1
        self.E2 = E2
        self.G12 = G12
        self.nu12 = nu12
        self.rho = rho
        self.name = name

    @classmethod
    def listFromPreCompFile(cls, fname):
        """initialize the object by extracting materials from a PreComp materials file

        Parameters
        ----------
        fname : str
            name of input file

        Returns
        -------
        materials : List(:class:`Orthotropic2DMaterial`)
            a list of all materials gathered from the file.

        """

        f = open(fname)

        skipLines(f, 3)

        materials = []
        for line in f:
            array = line.split()
            mat = cls(float(array[1]), float(array[2]), float(array[3]), float(array[4]), float(array[5]), array[6])

            materials.append(mat)
        f.close()

        return materials



import unittest

import numpy as np
import numpy.testing as npt

import wisdem.drivetrainse.drive_components as dc


class TestComponents(unittest.TestCase):
    def testBearing(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.MainBearing()

        discrete_inputs["bearing_type"] = "carb"
        inputs["D_bearing"] = 2.0
        inputs["D_shaft"] = 3.0
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(
            outputs["mb_I"] / outputs["mb_mass"], 0.125 * np.r_[2 * (4 * 1.5**2 + 3), (4 * 1.5**2 + 5) * np.ones(2)]
        )
        npt.assert_almost_equal(outputs["mb_mass"], (1 + 80 / 27) * 1561.4 * 3**2.6007)
        npt.assert_almost_equal(outputs["mb_max_defl_ang"], 0.5 * np.pi / 180)

        # Other valid types
        discrete_inputs["bearing_type"] = "crb"
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        discrete_inputs["bearing_type"] = "srb"
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        discrete_inputs["bearing_type"] = "trb"
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        # Invalid type
        try:
            discrete_inputs["bearing_type"] = 1
            myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        except ValueError:
            self.assertTrue(True)
        # Unknown type
        try:
            discrete_inputs["bearing_type"] = "trb1"
            myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        except ValueError:
            self.assertTrue(True)

    def testBrake(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.Brake(direct_drive=True)

        discrete_inputs["direct_drive"] = True
        inputs["rotor_diameter"] = 200.0
        inputs["rated_torque"] = 10e6
        inputs["s_rotor"] = 3.0
        inputs["s_gearbox"] = 0.0
        inputs["brake_mass_user"] = 0.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["brake_mass"], 12200)
        npt.assert_equal(outputs["brake_cm"], 3)
        npt.assert_equal(outputs["brake_I"], 12200 * np.r_[0.5, 0.25, 0.25])

        discrete_inputs["direct_drive"] = False
        inputs["s_gearbox"] = np.array([ 5.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["brake_mass"], 12200)
        npt.assert_equal(outputs["brake_cm"], 3)
        npt.assert_equal(outputs["brake_I"], 12200 * np.r_[0.5, 0.25, 0.25])

        inputs["brake_mass_user"] = np.array([ 42.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["brake_mass"], 42.0)
        npt.assert_equal(outputs["brake_cm"], 3)
        npt.assert_equal(outputs["brake_I"], 42.0 * np.r_[0.5, 0.25, 0.25])

    def testRPM_In(self):
        inputs = {}
        outputs = {}
        myobj = dc.RPM_Input(n_pc=20)

        inputs["minimum_rpm"] = 2.0
        inputs["rated_rpm"] = 10.0
        inputs["gear_ratio"] = 100.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        myobj.compute(inputs, outputs)

        x = np.linspace(2, 10, 20)
        npt.assert_equal(outputs["lss_rpm"], x)
        npt.assert_equal(outputs["hss_rpm"], 100 * x)

    def testGeneratorSimple(self):
        inputs = {}
        outputs = {}
        myobj = dc.GeneratorSimple(direct_drive=True, n_pc=20)

        inputs["rotor_diameter"] = 200.0
        inputs["machine_rating"] = 10e3
        inputs["rated_torque"] = 10e6
        inputs["L_generator"] = 3.6 * 1.5
        inputs["generator_mass_user"] = 0.0
        inputs["generator_radius_user"] = 0.0
        inputs["generator_efficiency_user"] = 0.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        inputs["lss_rpm"] = x = np.linspace(0.1, 10.0, 20)
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["R_generator"], 1.5)
        m = 37.68 * 10e3
        npt.assert_equal(outputs["generator_mass"], m)
        npt.assert_equal(outputs["generator_rotor_mass"], 0.5 * m)
        npt.assert_equal(outputs["generator_stator_mass"], 0.5 * m)
        npt.assert_equal(
            outputs["generator_I"], m * np.r_[0.5 * 1.5**2, (3 * 1.5**2 + (3.6 * 1.5) ** 2) / 12 * np.ones(2)]
        )
        npt.assert_equal(outputs["generator_rotor_I"], 0.5 * outputs["generator_I"])
        npt.assert_equal(outputs["generator_stator_I"], 0.5 * outputs["generator_I"])

        eff = 1.0 - (0.01007 / x * x[-1] + 0.02 + 0.06899 * x / x[-1])
        eff = np.maximum(1e-3, eff)
        npt.assert_almost_equal(outputs["generator_efficiency"], eff)

        myobj = dc.GeneratorSimple(direct_drive=False)
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["R_generator"], 1.5)
        m = np.mean([6.4737, 10.51, 5.34]) * 10e3**0.9223
        npt.assert_equal(outputs["generator_mass"], m)
        npt.assert_equal(outputs["generator_rotor_mass"], 0.5 * m)
        npt.assert_equal(outputs["generator_stator_mass"], 0.5 * m)
        npt.assert_equal(
            outputs["generator_I"], m * np.r_[0.5 * 1.5**2, (3 * 1.5**2 + (3.6 * 1.5) ** 2) / 12 * np.ones(2)]
        )
        npt.assert_equal(outputs["generator_rotor_I"], 0.5 * outputs["generator_I"])
        npt.assert_equal(outputs["generator_stator_I"], 0.5 * outputs["generator_I"])

        eff = 1.0 - (0.01289 / x * x[-1] + 0.0851 + 0.0 * x / x[-1])
        eff = np.maximum(1e-3, eff)
        npt.assert_almost_equal(outputs["generator_efficiency"], eff)

        eff = np.linspace(0.5, 1.0, 20)
        inputs["generator_efficiency_user"] = np.c_[x, eff]
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["generator_efficiency"], eff)

        inputs["generator_mass_user"] = np.array([ 2.0 ])
        inputs["generator_radius_user"] = np.array([ 3.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["R_generator"], 3.0)
        npt.assert_equal(outputs["generator_mass"], 2.0)
        npt.assert_equal(outputs["generator_rotor_mass"], 1.0)
        npt.assert_equal(outputs["generator_stator_mass"], 1.0)
        npt.assert_equal(outputs["generator_rotor_I"], 0.5 * outputs["generator_I"])
        npt.assert_equal(outputs["generator_stator_I"], 0.5 * outputs["generator_I"])

    def testElectronics(self):
        inputs = {}
        outputs = {}
        myobj = dc.Electronics()

        inputs["rotor_diameter"] = 200.0
        inputs["machine_rating"] = 10e3
        inputs["D_top"] = 5.0
        inputs["converter_mass_user"] = 0.0
        inputs["transformer_mass_user"] = 0.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        myobj.compute(inputs, outputs)
        s = 0.015 * 200
        m = np.mean([740.0, 817.5]) * 10 + np.mean([101.37, 503.83])
        npt.assert_almost_equal(outputs["converter_mass"], m)
        npt.assert_equal(outputs["converter_cm"], np.r_[0.0, -(2.5 + 0.5 * s), 0.5 * s])
        npt.assert_almost_equal(outputs["converter_I"], (1.0 / 6.0) * m * s**2)

        m = 1915 * 10 + 1910.0
        npt.assert_equal(outputs["transformer_mass"], m)
        npt.assert_equal(outputs["transformer_cm"], np.r_[0.0, -(2.5 + 0.5 * s), 0.5 * s])
        npt.assert_almost_equal(outputs["transformer_I"], (1.0 / 6.0) * m * s**2)

        inputs["converter_mass_user"] = np.array([ 42.0 ])
        inputs["transformer_mass_user"] = np.array([ 420.0 ])
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["converter_mass"], 42.0)
        npt.assert_equal(outputs["converter_cm"], np.r_[0.0, -(2.5 + 0.5 * s), 0.5 * s])
        npt.assert_almost_equal(outputs["converter_I"], (1.0 / 6.0) * 42 * s**2)
        npt.assert_equal(outputs["transformer_mass"], 420.0)
        npt.assert_equal(outputs["transformer_cm"], np.r_[0.0, -(2.5 + 0.5 * s), 0.5 * s])
        npt.assert_almost_equal(outputs["transformer_I"], (1.0 / 6.0) * 420 * s**2)

    def testYaw(self):
        inputs = {}
        outputs = {}
        myobj = dc.YawSystem()

        inputs["rotor_diameter"] = 200.0
        inputs["machine_rating"] = 10e3
        inputs["D_top"] = 5.0
        inputs["rho"] = 5e3
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["yaw_mass"], 5e3 * np.pi * 0.1 * 5**2 * 0.2 + 190 * 12)
        npt.assert_equal(outputs["yaw_cm"], 0.0)
        npt.assert_equal(outputs["yaw_I"], 0.0)

    def testMiscDirect(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.MiscNacelleComponents(direct_drive=True)

        discrete_inputs["upwind"] = False
        inputs["machine_rating"] = 10e3
        inputs["L_bedplate"] = 5.0
        inputs["H_bedplate"] = 4.0
        inputs["D_top"] = 6.0
        inputs["R_generator"] = 2.0
        inputs["overhang"] = 10.0
        inputs["generator_cm"] = 6.0
        inputs["rho_fiberglass"] = 2e3
        inputs["rho_castiron"] = 3e3
        inputs["hvac_mass_coeff"] = 0.1
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)

        L = 1.1 * 5  # (10 + 6.0)
        W = 1.1 * 2 * 2
        H = 1.1 * (2 + 4)
        npt.assert_equal(outputs["cover_mass"], 0.02 * 2e3 * 2 * (L * W + L * H + W * H))
        npt.assert_equal(outputs["cover_cm"], np.array([0.5 * (L - 5), 0.0, 0.5 * H]))

        npt.assert_equal(outputs["hvac_mass"], 0.1 * 10e3 * 2 * np.pi * 0.75 * 2)
        npt.assert_equal(outputs["hvac_cm"], 6.0)
        npt.assert_equal(outputs["hvac_I"], outputs["hvac_mass"] * 1.5**2 * np.r_[1.0, 0.5, 0.5])

        t = 0.04
        npt.assert_equal(outputs["platform_mass"], t * 3e3 * 12**2)
        npt.assert_equal(outputs["platform_cm"], 0.0)
        npt.assert_equal(
            outputs["platform_I"],
            outputs["platform_mass"] * np.array([t**2 + 12**2, t**2 + 12**2, 2 * 12**2]) / 12.0,
        )

        discrete_inputs["upwind"] = True
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["cover_mass"], 0.02 * 2e3 * 2 * (L * W + L * H + W * H))
        npt.assert_equal(outputs["cover_cm"], np.array([-0.5 * (L - 5), 0.0, 0.5 * H]))

        npt.assert_equal(outputs["hvac_mass"], 0.1 * 10e3 * 2 * np.pi * 0.75 * 2)
        npt.assert_equal(outputs["hvac_cm"], 6.0)
        npt.assert_equal(outputs["hvac_I"], outputs["hvac_mass"] * 1.5**2 * np.r_[1.0, 0.5, 0.5])

        npt.assert_equal(outputs["platform_mass"], t * 3e3 * 12**2)
        npt.assert_equal(outputs["platform_cm"], 0.0)
        npt.assert_equal(
            outputs["platform_I"],
            outputs["platform_mass"] * np.array([t**2 + 12**2, t**2 + 12**2, 2 * 12**2]) / 12.0,
        )

    def testMiscGeared(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.MiscNacelleComponents(direct_drive=False)

        discrete_inputs["upwind"] = False
        inputs["machine_rating"] = 10e3
        inputs["L_bedplate"] = 5.0
        inputs["H_bedplate"] = 4.0
        inputs["D_top"] = 6.0
        inputs["R_generator"] = 2.0
        inputs["overhang"] = 10.0
        inputs["generator_cm"] = 6.0
        inputs["rho_fiberglass"] = 2e3
        inputs["rho_castiron"] = 3e3
        inputs["hvac_mass_coeff"] = 0.1
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)

        L = 1.1 * (10 + 6.0)
        W = 1.1 * 2 * 2
        H = 1.1 * (2 + 4)
        npt.assert_equal(outputs["cover_mass"], 0.02 * 2e3 * 2 * (L * W + L * H + W * H))
        npt.assert_equal(outputs["cover_cm"], np.array([0.5 * (L - 5), 0.0, 0.5 * H]))

        npt.assert_equal(outputs["hvac_mass"], 0.1 * 10e3 * 2 * np.pi * 0.75 * 2)
        npt.assert_equal(outputs["hvac_cm"], 6.0)
        npt.assert_equal(outputs["hvac_I"], outputs["hvac_mass"] * 1.5**2 * np.r_[1.0, 0.5, 0.5])

        t = 0.04
        npt.assert_equal(outputs["platform_mass"], t * 3e3 * L * W)
        npt.assert_equal(outputs["platform_cm"], 0.0)
        npt.assert_equal(
            outputs["platform_I"],
            outputs["platform_mass"] * np.array([t**2 + W**2, t**2 + L**2, L**2 + W**2]) / 12.0,
        )

        discrete_inputs["upwind"] = True
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["cover_mass"], 0.02 * 2e3 * 2 * (L * W + L * H + W * H))
        npt.assert_equal(outputs["cover_cm"], np.array([-0.5 * (L - 5), 0.0, 0.5 * H]))

        npt.assert_equal(outputs["hvac_mass"], 0.1 * 10e3 * 2 * np.pi * 0.75 * 2)
        npt.assert_equal(outputs["hvac_cm"], 6.0)
        npt.assert_equal(outputs["hvac_I"], outputs["hvac_mass"] * 1.5**2 * np.r_[1.0, 0.5, 0.5])

        npt.assert_equal(outputs["platform_mass"], t * 3e3 * L * W)
        npt.assert_equal(outputs["platform_cm"], 0.0)
        npt.assert_equal(
            outputs["platform_I"],
            outputs["platform_mass"] * np.array([t**2 + W**2, t**2 + L**2, L**2 + W**2]) / 12.0,
        )

    def testNacelle_noTilt(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.NacelleSystemAdder()

        discrete_inputs["upwind"] = True
        discrete_inputs["uptower"] = True
        inputs["tilt"] = 0.0
        inputs["constr_height"] = 2.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        inputs["x_bedplate"] = -2 * np.ones(5)
        components = [
            "mb1",
            "mb2",
            "lss",
            "hss",
            "gearbox",
            "generator",
            "generator_stator",
            "generator_rotor",
            "hvac",
            "brake",
            "nose",
            "bedplate",
            "platform",
            "yaw",
            "cover",
            "transformer",
            "converter",
        ]
        cm3 = ["gearbox", "transformer", "converter", "yaw", "bedplate", "platform", "cover"]
        for k in components:
            inputs[k + "_mass"] = np.array([ 1e3 ])
            inputs[k + "_I"] = 1e3 * np.array([1, 2, 3])
            if k in cm3:
                inputs[k + "_cm"] = np.array([-5.0, 0.0, 2.0])
            else:
                inputs[k + "_cm"] = np.array([ 3.0 ])

        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["other_mass"], 1e3 * 6)
        npt.assert_equal(outputs["nacelle_mass"], 1e3 * (len(components) - 2))  # gen stator / rotor duplication
        npt.assert_equal(outputs["nacelle_cm"], np.r_[-5.0, 0.0, 2.0])
        npt.assert_equal(outputs["nacelle_I"], 1e3 * (len(components) - 2) * np.r_[1.0, 2.0, 3.0, np.zeros(3)])

        discrete_inputs["upwind"] = False
        for k in cm3:
            inputs[k + "_cm"][0] *= -1.0
        inputs["x_bedplate"] *= -1.0
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["other_mass"], 1e3 * 6)
        npt.assert_equal(outputs["nacelle_mass"], 1e3 * (len(components) - 2))
        npt.assert_equal(outputs["nacelle_cm"], np.r_[5.0, 0.0, 2.0])
        npt.assert_equal(outputs["nacelle_I"], 1e3 * (len(components) - 2) * np.r_[1.0, 2.0, 3.0, np.zeros(3)])

        discrete_inputs["uptower"] = False
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["other_mass"], 1e3 * 6)
        npt.assert_equal(outputs["nacelle_mass"], 1e3 * (len(components) - 4))
        npt.assert_equal(outputs["nacelle_cm"], np.r_[5.0, 0.0, 2.0])
        npt.assert_equal(outputs["nacelle_I"], 1e3 * (len(components) - 4) * np.r_[1.0, 2.0, 3.0, np.zeros(3)])

    def testNacelle_withTilt(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.NacelleSystemAdder()

        discrete_inputs["upwind"] = True
        discrete_inputs["uptower"] = True
        inputs["tilt"] = 5.0
        tr = 5 * np.pi / 180.0
        inputs["constr_height"] = 2.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        inputs["x_bedplate"] = -2 * np.ones(5)
        components = [
            "mb1",
            "mb2",
            "lss",
            "hss",
            "gearbox",
            "generator",
            "generator_stator",
            "generator_rotor",
            "hvac",
            "brake",
            "nose",
            "bedplate",
            "platform",
            "yaw",
            "cover",
            "transformer",
            "converter",
        ]
        cm3 = ["gearbox", "transformer", "converter", "yaw", "bedplate", "platform", "cover"]
        for k in components:
            inputs[k + "_mass"] = np.array([ 1e3 ])
            inputs[k + "_I"] = 1e3 * np.array([1, 2, 3])
            if k in cm3:
                inputs[k + "_cm"] = np.array([-3.0 * np.cos(tr) - 2, 0.0, 2 + 3.0 * np.sin(tr)])
            else:
                inputs[k + "_cm"] = np.array([ 3.0 ])

        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["other_mass"], 1e3 * 6)
        npt.assert_equal(outputs["nacelle_mass"], 1e3 * (len(components) - 2))  # gen stator / rotor duplication
        npt.assert_almost_equal(outputs["nacelle_cm"], np.r_[-3.0 * np.cos(tr) - 2, 0.0, 2 + 3.0 * np.sin(tr)])
        # npt.assert_equal(outputs['nacelle_I'], 1e3*len(components)*np.r_[1.0, 2.0, 3.0, np.zeros(3)])

        discrete_inputs["upwind"] = False
        for k in cm3:
            inputs[k + "_cm"][0] *= -1.0
        inputs["x_bedplate"] *= -1.0
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["other_mass"], 1e3 * 6)
        npt.assert_equal(outputs["nacelle_mass"], 1e3 * (len(components) - 2))
        npt.assert_almost_equal(outputs["nacelle_cm"], np.r_[3.0 * np.cos(tr) + 2, 0.0, 2 + 3.0 * np.sin(tr)])
        # npt.assert_equal(outputs['nacelle_I'], 1e3*len(components)*np.r_[1.0, 2.0, 3.0, np.zeros(3)])

        discrete_inputs["uptower"] = False
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["other_mass"], 1e3 * 6)
        npt.assert_equal(outputs["nacelle_mass"], 1e3 * (len(components) - 4))
        npt.assert_almost_equal(outputs["nacelle_cm"], np.r_[3.0 * np.cos(tr) + 2, 0.0, 2 + 3.0 * np.sin(tr)])

    def testRNA(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}
        myobj = dc.RNA_Adder()

        discrete_inputs["upwind"] = True
        inputs["tilt"] = 0.0
        inputs["L_drive"] = 10.0
        inputs["blades_mass"] = 100e3
        inputs["nacelle_mass"] = 200e3
        inputs["hub_system_mass"] = 25e3
        inputs["hub_system_cm"] = 2.0
        inputs["blades_cm"] = 0.0
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        inputs["shaft_start"] = np.zeros(3)
        inputs["blades_I"] = 100e3 * np.arange(1, 7)
        inputs["nacelle_I_TT"] = 200e3 * np.arange(1, 7)
        inputs["nacelle_cm"] = np.array([-5.0, 0.0, 0.0])
        inputs["hub_system_I"] = 25e3 * np.arange(1, 7)

        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["rotor_mass"], 125e3)
        npt.assert_equal(outputs["rna_mass"], 325e3)
        npt.assert_equal(outputs["rna_cm"], np.r_[(-125 * 12 - 200 * 5) / 325, 0.0, 0.0])
        I0 = 325e3 * np.arange(1, 7)
        I0[1:3] += 125e3 * 12**2
        npt.assert_equal(outputs["rna_I_TT"], I0)

        discrete_inputs["upwind"] = False
        inputs["nacelle_cm"] = np.array([5.0, 0.0, 0.0])
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["rotor_mass"], 125e3)
        npt.assert_equal(outputs["rna_mass"], 325e3)
        npt.assert_equal(outputs["rna_cm"], np.r_[(125 * 12 + 200 * 5) / 325, 0.0, 0.0])
        npt.assert_equal(outputs["rna_I_TT"], I0)

        inputs["tilt"] = np.array([ 5.0 ])
        tr = 5 * np.pi / 180.0
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        npt.assert_equal(outputs["rotor_mass"], 125e3)
        npt.assert_equal(outputs["rna_mass"], 325e3)
        npt.assert_almost_equal(
            outputs["rna_cm"], np.r_[(125 * 12 * np.cos(tr) + 200 * 5) / 325, 0.0, 125 * 12 * np.sin(tr) / 325]
        )

    def testDynamics(self):
        inputs = {}
        outputs = {}
        myobj = dc.DriveDynamics()

        inputs["lss_spring_constant"] = 2.0
        inputs["hss_spring_constant"] = 3.0
        inputs["gear_ratio"] = 1.0
        inputs["damping_ratio"] = 0.5
        for k in inputs:
            inputs[k] = np.array( [inputs[k]] )
        inputs["blades_I"] = 30 * np.ones(6)
        inputs["hub_system_I"] = 20 * np.ones(6)
        
        myobj.compute(inputs, outputs)
        npt.assert_equal(outputs["drivetrain_spring_constant"], 2.0)
        npt.assert_equal(outputs["drivetrain_damping_coefficient"], 10)

        inputs["gear_ratio"] = 5.0
        myobj.compute(inputs, outputs)
        npt.assert_almost_equal(outputs["drivetrain_spring_constant"], 150.0 / (2 + 75))
        npt.assert_almost_equal(outputs["drivetrain_damping_coefficient"], np.sqrt(50 * 150.0 / (2 + 75)))


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pathlib import Path

import numpy as np
#import numpy.testing as npt

import wisdem.inputs.validation as val

sample_yaml = os.path.join(os.path.dirname(os.path.realpath(__file__)), "sample.yaml")

class TestValidation(unittest.TestCase):

    def testLoadStringPath(self):
        obj_str = val.load_yaml(val.fschema_opt)
        obj_path = val.load_yaml(Path(val.fschema_opt))

        self.assertEqual(obj_str, obj_path)

    def test_write_dict(self):
        obj1 = val.load_yaml(sample_yaml)
        obj2 = val.load_yaml(sample_yaml)

        obj1['components']['tower']['outer_shape_bem']['reference_axis']['z']['values'] = list(
            obj1['components']['tower']['outer_shape_bem']['reference_axis']['z']['values']
            )
        
        obj2['components']['tower']['outer_shape_bem']['reference_axis']['z']['values'] = np.array(
            obj2['components']['tower']['outer_shape_bem']['reference_axis']['z']['values']
            )

        ftemp1 = 'temp1.yaml'
        ftemp2 = 'temp2.yaml'
        val.write_yaml(obj1, ftemp1)
        val.write_yaml(obj2, ftemp2)
        
        obj1p = val.load_yaml(ftemp1)
        obj2p = val.load_yaml(ftemp2)

        self.assertEqual(obj1p, obj2p)
        
if __name__ == "__main__":
    unittest.main()



"""Tests for the `MonopileInstallation` class without feeder barges."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

import pandas as pd
import pytest

from wisdem.orbit import ProjectManager
from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.phases.install import MonopileInstallation
from wisdem.test.test_orbit.data import test_weather

config_wtiv = extract_library_specs("config", "single_wtiv_mono_install")
config_wtiv_feeder = extract_library_specs("config", "multi_wtiv_mono_install")
config_wtiv_multi_feeder = deepcopy(config_wtiv_feeder)
config_wtiv_multi_feeder["num_feeders"] = 2


@pytest.mark.parametrize(
    "config",
    (config_wtiv, config_wtiv_feeder, config_wtiv_multi_feeder),
    ids=["wtiv_only", "single_feeder", "multi_feeder"],
)
def test_simulation_setup(config):
    sim = MonopileInstallation(config)
    assert sim.config == config
    assert sim.env
    assert sim.port
    assert sim.port.crane.capacity == config["port"]["num_cranes"]
    assert sim.num_monopiles == config["plant"]["num_turbines"]

    mp = len([i for i in sim.port.items if i.type == "Monopile"])
    assert mp == sim.num_monopiles

    tp = len([i for i in sim.port.items if i.type == "TransitionPiece"])
    assert tp == sim.num_monopiles


@pytest.mark.parametrize(
    "config",
    (config_wtiv, config_wtiv_feeder, config_wtiv_multi_feeder),
    ids=["wtiv_only", "single_feeder", "multi_feeder"],
)
def test_vessel_initialization(config):
    sim = MonopileInstallation(config)
    assert sim.wtiv
    assert sim.wtiv.jacksys
    assert sim.wtiv.crane
    assert sim.wtiv.storage

    if config.get("feeder", None) is not None:
        assert len(sim.feeders) == config["num_feeders"]

        for feeder in sim.feeders:
            assert feeder.jacksys
            assert feeder.storage


@pytest.mark.parametrize(
    "config",
    (config_wtiv, config_wtiv_feeder, config_wtiv_multi_feeder),
    ids=["wtiv_only", "single_feeder", "multi_feeder"],
)
@pytest.mark.parametrize("weather", (None, test_weather), ids=["no_weather", "test_weather"])
def test_for_complete_logging(weather, config):
    sim = MonopileInstallation(config, weather=weather)
    sim.run()

    df = pd.DataFrame(sim.env.actions)
    df = df.assign(shift=(df["time"] - df["time"].shift(1)))

    for vessel in df["agent"].unique():
        _df = df[df["agent"] == vessel].copy()
        _df = _df.assign(shift=(_df["time"] - _df["time"].shift(1)))
        assert (_df["shift"] - _df["duration"]).abs().max() < 1e-9

    assert ~df["cost"].isnull().any()
    _ = sim.agent_efficiencies
    _ = sim.detailed_output


def test_kwargs():
    sim = MonopileInstallation(config_wtiv)
    sim.run()
    baseline = sim.total_phase_time

    keywords = [
        "mono_embed_len",
        "mono_drive_rate",
        "mono_fasten_time",
        "mono_release_time",
        "tp_fasten_time",
        "tp_release_time",
        "tp_bolt_time",
        "site_position_time",
        "crane_reequip_time",
        "rov_survey_time",
    ]

    failed = []

    for kw in keywords:
        default = pt[kw]

        if kw == "mono_drive_rate":
            _new = default - 2

            if _new <= 0:
                raise Exception(f"'{kw}' is less than 0.")

            kwargs = {kw: _new}

        else:
            kwargs = {kw: default + 2}

        new_sim = MonopileInstallation(config_wtiv, **kwargs)
        new_sim.run()
        new_time = new_sim.total_phase_time

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"'{failed}' not affecting results.")

    else:
        assert True


def test_kwargs_in_ProjectManager():
    base = deepcopy(config_wtiv)
    base["install_phases"] = ["MonopileInstallation"]
    project = ProjectManager(base)
    project.run()
    baseline = project.phase_times["MonopileInstallation"]

    keywords = [
        "mono_embed_len",
        "mono_drive_rate",
        "mono_fasten_time",
        "mono_release_time",
        "tp_fasten_time",
        "tp_release_time",
        "tp_bolt_time",
        "site_position_time",
        "crane_reequip_time",
        "rov_survey_time",
    ]

    failed = []

    for kw in keywords:
        default = pt[kw]

        if kw == "mono_drive_rate":
            _new = default - 2

            if _new <= 0:
                raise Exception(f"'{kw}' is less than 0.")

            processes = {kw: _new}

        else:
            processes = {kw: default + 2}

        new_config = deepcopy(base)
        new_config["processes"] = processes

        new_project = ProjectManager(new_config)
        new_project.run()
        new_time = new_project.phase_times["MonopileInstallation"]

        if new_time > baseline:
            pass

        else:
            failed.append(kw)

    if failed:
        raise Exception(f"'{failed}' not affecting results.")

    else:
        assert True


def test_grout_kwargs():
    sim = MonopileInstallation(config_wtiv)
    sim.run()

    assert "Bolt TP" in list([a["action"] for a in sim.env.actions])

    sim = MonopileInstallation(config_wtiv, tp_connection_type="grouted")
    sim.run()

    assert "Pump TP Grout" in list([a["action"] for a in sim.env.actions])
    assert "Cure TP Grout" in list([a["action"] for a in sim.env.actions])



"""
Testing framework for common monopile installation tasks.
"""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


import pytest

from wisdem.orbit.core.exceptions import MissingComponent
from wisdem.orbit.phases.install.monopile_install.common import (
    drive_monopile,
    lower_monopile,
    upend_monopile,
    bolt_transition_piece,
    lower_transition_piece,
    cure_transition_piece_grout,
    pump_transition_piece_grout,
)


@pytest.mark.parametrize(
    "task, log, args",
    [
        (upend_monopile, "Upend Monopile", [100]),
        (lower_monopile, "Lower Monopile", []),
        (drive_monopile, "Drive Monopile", []),
        (lower_transition_piece, "Lower TP", []),
        (bolt_transition_piece, "Bolt TP", []),
        (pump_transition_piece_grout, "Pump TP Grout", []),
        (cure_transition_piece_grout, "Cure TP Grout", []),
    ],
)
def test_task(env, wtiv, task, log, args):
    env.register(wtiv)
    wtiv.initialize(mobilize=False)

    task(wtiv, *args, site_depth=10)
    env.run()

    actions = [a["action"] for a in env.actions]
    assert log in actions


@pytest.mark.parametrize(
    "task, log, args",
    [
        (upend_monopile, "Upend Monopile", [100]),
        (lower_monopile, "Lower Monopile", []),
        (drive_monopile, "Drive Monopile", []),
    ],
)
def test_task_fails(env, feeder, task, log, args):
    env.register(feeder)
    feeder.initialize(mobilize=False)

    with pytest.raises(MissingComponent):
        task(feeder, *args, site_depth=10)
        env.run()

    actions = [a["action"] for a in env.actions]
    assert log not in actions



#!/usr/bin/env python3
import os

from wisdem import run_wisdem

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_wt_input = mydir + os.sep + "nrel5mw_tower.yaml"
fname_modeling_options = mydir + os.sep + "modeling_options.yaml"
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"

wt_opt, analysis_options, opt_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

# print results from the analysis or optimization
z = 0.5 * (wt_opt["towerse.z_full"][:-1] + wt_opt["towerse.z_full"][1:])
print("zs =", wt_opt["towerse.z_full"])
print("ds =", wt_opt["towerse.outer_diameter_full"])
print("ts =", wt_opt["towerse.t_full"])
print("mass (kg) =", wt_opt["towerse.tower_mass"])
print("cg (m) =", wt_opt["towerse.tower_center_of_mass"])
print("d:t constraint =", wt_opt["towerse.constr_d_to_t"])
print("taper ratio constraint =", wt_opt["towerse.constr_taper"])
print("\nwind: ", wt_opt["towerse.env1.Uref"], wt_opt["towerse.env2.Uref"])
print("freq (Hz) =", wt_opt["towerse.tower.structural_frequencies"])
print("Fore-aft mode shapes =", wt_opt["towerse.tower.fore_aft_modes"])
print("Side-side mode shapes =", wt_opt["towerse.tower.side_side_modes"])
print("top_deflection1 (m) =", wt_opt["towerse.tower.top_deflection"])
print("Tower base forces1 (N) =", wt_opt["towerse.tower.turbine_F"])
print("Tower base moments1 (Nm) =", wt_opt["towerse.tower.turbine_M"])
print("stress1 =", wt_opt["towerse.post.constr_stress"])
print("GL buckling =", wt_opt["towerse.post.constr_global_buckling"])
print("Shell buckling =", wt_opt["towerse.post.constr_shell_buckling"])
